"""
His Maj. King of Spain Contest, CW
Status:	            Active
Geographic Focus:	Worldwide
Participation:	    Worldwide
Awards:	            Worldwide
Mode:	            CW
Bands:	            160, 80, 40, 20, 15, 10m
Classes:	        Single Op All Band (QRP/Low/High)
                    Single Op All Band Youth
                    Single Op Single Band
                    Multi-Op (Low/High)
Max power:	        HP: >100 watts
                    LP: 100 watts
                    QRP: 5 watts
Exchange:	        EA: RST + province
                    non-EA: RST + Serial No.
Work stations:	    Once per band
QSO Points:	        (see rules)
Multipliers:	    Each EA province once per band
                    Each EADX100 entity once per band
                    Each special (EA0) station once per band
Score Calculation:	Total score = total QSO points x total mults
E-mail logs to:	    (none)
Upload log at:	    https://concursos.ure.es/en/logs/
Mail logs to:	    (none)
Find rules at:	    https://concursos.ure.es/en/s-m-el-rey-de-espana-cw/bases/
Cabrillo name:	    EA-MAJESTAD-CW
"""

# pylint: disable=invalid-name, unused-argument, unused-variable, c-extension-no-member

# EA1: AV, BU, C, LE, LO, LU, O, OU, P, PO, S, SA, SG, SO, VA, ZA
# EA2: BI, HU, NA, SS, TE, VI, Z
# EA3: B, GI, L, T
# EA4: BA, CC, CR, CU, GU, M, TO
# EA5: A, AB, CS, MU, V
# EA6: IB
# EA7: AL, CA, CO, GR, H, J, MA, SE
# EA8: GC, TF
# EA9: CE, ML


import datetime
import logging

from pathlib import Path
from PyQt6 import QtWidgets

from not1mm.lib.ham_utility import get_logged_band
from not1mm.lib.plugin_common import gen_adif, imp_adif, get_points, online_score_xml
from not1mm.lib.version import __version__

logger = logging.getLogger(__name__)

EXCHANGE_HINT = "Province or #"

name = "His Maj. King of Spain Contest, CW"
mode = "CW"  # CW SSB BOTH RTTY
cabrillo_name = "EA-MAJESTAD-CW"

columns = [
    "YYYY-MM-DD HH:MM:SS",
    "Call",
    "Freq",
    "Snt",
    "Rcv",
    "SentNr",
    "RcvNr",
    "PTS",
]

advance_on_space = [True, True, True, True, True]

# 1 once per contest, 2 work each band, 3 each band/mode, 4 no dupe checking
dupe_type = 2


def init_contest(self) -> None:
    """setup plugin"""
    set_tab_next(self)
    set_tab_prev(self)
    interface(self)
    self.next_field = self.other_2


def interface(self) -> None:
    """
    Setup the user interface.
    Unhides the input fields and sets the lebels.
    """
    self.field1.show()
    self.field2.show()
    self.field3.show()
    self.field4.show()
    self.snt_label.setText("SNT")
    self.field1.setAccessibleName("RST Sent")
    self.other_label.setText("SentNR")
    self.field3.setAccessibleName("Sent Number")
    self.exch_label.setText("Prov or SN")
    self.field4.setAccessibleName("Province or Serial Number")


def reset_label(self) -> None:
    """
    Reset label after field cleared.
    Not needed for this contest.
    """


def set_tab_next(self) -> None:
    """
    Set TAB Advances.
    Defines which which of the fields are next to get focus when the TAB key is pressed.
    """
    self.tab_next = {
        self.callsign: self.sent,
        self.sent: self.receive,
        self.receive: self.other_1,
        self.other_1: self.other_2,
        self.other_2: self.callsign,
    }


def set_tab_prev(self) -> None:
    """
    Set TAB Advances.
    Defines which which of the fields are next to get focus when the Shift-TAB key is pressed.
    """
    self.tab_prev = {
        self.callsign: self.other_2,
        self.sent: self.callsign,
        self.receive: self.sent,
        self.other_1: self.receive,
        self.other_2: self.other_1,
    }


def validate(self) -> bool:
    """Not Used"""
    return True


def set_contact_vars(self) -> None:
    """Contest Specific"""
    self.contact["SNT"] = self.sent.text()
    self.contact["RCV"] = self.receive.text()
    self.contact["NR"] = self.other_2.text().upper()
    self.contact["SentNr"] = self.other_1.text()


def predupe(self) -> None:
    """called after callsign entered. Not needed here."""


def prefill(self) -> None:
    """
    Fill the SentNR field with either the next serial number or the province.
    """
    # result = self.database.get_serial()
    # serial_nr = str(result.get("serial_nr", "1")).zfill(3)
    serial_nr = str(self.current_sn).zfill(3)
    if serial_nr == "None":
        serial_nr = "001"

    exchange = self.contest_settings.get("SentExchange", "").replace("#", serial_nr)
    if len(self.other_1.text()) == 0:
        self.other_1.setText(exchange)


def points(self) -> int:
    """
    Calculate the points for this contact.
    """
    # EA: 2 points per QSO with EA
    # EA: 1 point per QSO with non-EA
    # non-EA: 3 points per QSO with EA
    # non-EA: 1 point per QSO with non-EA

    if self.contact_is_dupe > 0:
        return 0

    ea_prefixes = ["EA", "EA1", "EA2", "EA3", "EA4", "EA5", "EA6", "EA7", "EA8", "EA9"]

    me = None
    him = None

    result = self.cty_lookup(self.station.get("Call", ""))
    if result is not None:
        item = result.get(next(iter(result)))
        me = item.get("primary_pfx", "")

    result = self.cty_lookup(self.contact.get("Call", ""))
    if result is not None:
        item = result.get(next(iter(result)))
        him = item.get("primary_pfx", "")

    if me is not None and him is not None:
        if me in ea_prefixes and him in ea_prefixes:
            return 2
        elif me in ea_prefixes and him not in ea_prefixes:
            return 1
        elif me not in ea_prefixes and him in ea_prefixes:
            return 3
        else:
            return 1

    return 1


def show_mults(self, rtc=None) -> int:
    """Return display string for mults"""

    ea_provinces = 0
    # dx = 0
    ef0f = 0
    eadx100 = 0

    # Each EADX100 entity once per band
    sql = (
        "select count(DISTINCT(CountryPrefix || ':' || Band)) as mult_count "
        f"from dxlog where ContestNR = {self.database.current_contest};"
    )
    result = self.database.exec_sql(sql)
    if result:
        eadx100 = result.get("mult_count", 0)

    # Each EA province once per band
    sql = (
        "select count(DISTINCT(NR || ':' || Band)) as mult_count "
        f"from dxlog where ContestNR = {self.database.current_contest} and typeof(NR) = 'text';"
    )
    result = self.database.exec_sql(sql)
    if result:
        ea_provinces = result.get("mult_count", 0)

    # # Each USA, VE, JA or VK call area once per band
    # sql = (
    #     "select count(DISTINCT(CountryPrefix || ':' || substr(WPXPrefix, -1) || ':' || Band)) as mult_count "
    #     f"from dxlog where CountryPrefix in ('K', 'VE', 'VK', 'JA') and ContestNR = {self.database.current_contest};"
    # )
    # result = self.database.exec_sql(sql)
    # if result:
    #     dx = result.get("mult_count", 0)

    # Each QSO with EF0F/8 once per band
    sql = (
        "select count(DISTINCT(Band)) as mult_count "
        f"from dxlog where Call = 'EF0F/8' and ContestNR = {self.database.current_contest};"
    )
    result = self.database.exec_sql(sql)
    if result:
        ef0f = result.get("mult_count", 0)

    if rtc is not None:
        return 0, 0

    return ea_provinces + ef0f + eadx100


def show_qso(self) -> int:
    """Return qso count"""
    result = self.database.fetch_qso_count()
    if result:
        return int(result.get("qsos", 0))
    return 0


def calc_score(self) -> int:
    """Return calculated score"""
    _points = get_points(self)
    _mults = show_mults(self)

    return _points * _mults


def adif(self) -> None:
    """Call the generate ADIF function"""
    gen_adif(self, cabrillo_name, contest_id=cabrillo_name)


def output_cabrillo_line(line_to_output, ending, file_descriptor, file_encoding):
    """"""
    print(
        line_to_output.encode(file_encoding, errors="ignore").decode(),
        end=ending,
        file=file_descriptor,
    )


def cabrillo(self, file_encoding):
    """Generates Cabrillo file. Maybe."""
    # https://www.cqwpx.com/cabrillo.htm
    logger.debug("******Cabrillo*****")
    logger.debug("Station: %s", f"{self.station}")
    logger.debug("Contest: %s", f"{self.contest_settings}")
    now = datetime.datetime.now()
    date_time = now.strftime("%Y-%m-%d_%H-%M-%S")
    filename = (
        str(Path.home())
        + "/"
        + f"{self.station.get('Call', '').upper().replace('/','-')}_{cabrillo_name}_{date_time}.log"
    )
    logger.debug("%s", filename)
    log = self.database.fetch_all_contacts_asc()
    try:
        with open(filename, "w", encoding=file_encoding, newline="") as file_descriptor:
            output_cabrillo_line(
                "START-OF-LOG: 3.0",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CREATED-BY: Not1MM v{__version__}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CONTEST: {cabrillo_name}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.station.get("Club", ""):
                output_cabrillo_line(
                    f"CLUB: {self.station.get('Club', '')}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"CALLSIGN: {self.station.get('Call','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"LOCATION: {self.station.get('ARRLSection', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-OPERATOR: {self.contest_settings.get('OperatorCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-ASSISTED: {self.contest_settings.get('AssistedCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-BAND: {self.contest_settings.get('BandCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            mode = self.contest_settings.get("ModeCategory", "")
            if mode in ["SSB+CW", "SSB+CW+DIGITAL"]:
                mode = "MIXED"
            output_cabrillo_line(
                f"CATEGORY-MODE: {mode}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-TRANSMITTER: {self.contest_settings.get('TransmitterCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.contest_settings.get("OverlayCategory", "") != "N/A":
                output_cabrillo_line(
                    f"CATEGORY-OVERLAY: {self.contest_settings.get('OverlayCategory','')}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"GRID-LOCATOR: {self.station.get('GridSquare','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-POWER: {self.contest_settings.get('PowerCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )

            output_cabrillo_line(
                f"CLAIMED-SCORE: {calc_score(self)}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            ops = ""
            list_of_ops = self.database.get_ops()
            for op in list_of_ops:
                ops += f"{op.get('Operator', '')}, "
            if self.station.get("Call", "") not in ops:
                ops += f"@{self.station.get('Call','')}"
            else:
                ops = ops.rstrip(", ")
            output_cabrillo_line(
                f"OPERATORS: {ops}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"NAME: {self.station.get('Name', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('Street1', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-CITY: {self.station.get('City', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-STATE-PROVINCE: {self.station.get('State', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-POSTALCODE: {self.station.get('Zip', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-COUNTRY: {self.station.get('Country', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"EMAIL: {self.station.get('Email', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            for contact in log:
                the_date_and_time = contact.get("TS", "")
                themode = contact.get("Mode", "")
                if themode in ("CW-U", "CW-L", "CW-R", "CWR"):
                    themode = "CW"
                if themode == "LSB" or themode == "USB":
                    themode = "PH"
                if themode == "RTTY":
                    themode = "RY"
                frequency = str(round(contact.get("Freq", "0"))).rjust(5)

                loggeddate = the_date_and_time[:10]
                loggedtime = the_date_and_time[11:13] + the_date_and_time[14:16]
                output_cabrillo_line(
                    f"QSO: {frequency} {themode} {loggeddate} {loggedtime} "
                    f"{contact.get('StationPrefix', '').ljust(13)} "
                    f"{str(contact.get('SNT', '')).ljust(3)} "
                    f"{str(contact.get('SentNr', '')).ljust(6)} "
                    f"{contact.get('Call', '').ljust(13)} "
                    f"{str(contact.get('RCV', '')).ljust(3)} "
                    f"{str(contact.get('NR', '')).ljust(6)}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line("END-OF-LOG:", "\r\n", file_descriptor, file_encoding)
        self.show_message_box(f"Cabrillo saved to: {filename}")
    except IOError as exception:
        logger.critical("cabrillo: IO error: %s, writing to %s", exception, filename)
        self.show_message_box(f"Error saving Cabrillo: {exception} {filename}")
        return


def recalculate_mults(self) -> None:
    """Recalculates multipliers after change in logged qso."""


def process_esm(self, new_focused_widget=None, with_enter=False):
    """ESM State Machine"""

    # self.pref["run_state"]

    # -----===== Assigned F-Keys =====-----
    # self.esm_dict["CQ"]
    # self.esm_dict["EXCH"]
    # self.esm_dict["QRZ"]
    # self.esm_dict["AGN"]
    # self.esm_dict["HISCALL"]
    # self.esm_dict["MYCALL"]
    # self.esm_dict["QSOB4"]

    # ----==== text fields ====----
    # self.callsign
    # self.sent
    # self.receive
    # self.other_1
    # self.other_2

    if new_focused_widget is not None:
        self.current_widget = self.inputs_dict.get(new_focused_widget)

    # print(f"checking esm {self.current_widget=} {with_enter=} {self.pref.get("run_state")=}")

    for a_button in [
        self.esm_dict["CQ"],
        self.esm_dict["EXCH"],
        self.esm_dict["QRZ"],
        self.esm_dict["AGN"],
        self.esm_dict["HISCALL"],
        self.esm_dict["MYCALL"],
        self.esm_dict["QSOB4"],
    ]:
        if a_button is not None:
            self.restore_button_color(a_button)

    buttons_to_send = []

    if self.pref.get("run_state"):
        if self.current_widget == "callsign":
            if len(self.callsign.text()) < 3:
                self.make_button_green(self.esm_dict["CQ"])
                buttons_to_send.append(self.esm_dict["CQ"])
            elif len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["HISCALL"])
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["HISCALL"])
                buttons_to_send.append(self.esm_dict["EXCH"])

        elif self.current_widget in ["other_2"]:
            if self.other_2.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["QRZ"])
                buttons_to_send.append(self.esm_dict["QRZ"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    if button == self.esm_dict["HISCALL"]:
                        self.process_function_key(button, rttysendrx=False)
                        continue
                    self.process_function_key(button)
    else:
        if self.current_widget == "callsign":
            if len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["MYCALL"])
                buttons_to_send.append(self.esm_dict["MYCALL"])

        elif self.current_widget in ["other_2"]:
            if self.other_2.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["EXCH"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    self.process_function_key(button)


def populate_history_info_line(self):
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        self.history_info.setText(
            f"{result.get('Call', '')}, {result.get('Name', '')}, {result.get('Exch1', '')}, {result.get('UserText','...')}"
        )
    else:
        self.history_info.setText("")


def check_call_history(self):
    """"""
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        self.history_info.setText(f"{result.get('UserText','')}")
        if self.other_2.text() == "":
            self.other_2.setText(f"{result.get('Exch1', '')}")


def get_mults(self):
    """"""
    mults = {}
    mults["country"], mults["state"] = show_mults(self, rtc=True)
    return mults


def just_points(self):
    """"""
    return get_points(self)


def set_self(the_outie):
    """..."""
    globals()["ALTEREGO"] = the_outie


def ft8_handler(the_packet: dict):
    print(f"{the_packet=}")
    """Process FT8 QSO packets
    # FT8
    # {
    #     'CALL': 'KE0OG',
    #     'GRIDSQUARE': 'DM10AT',
    #     'MODE': 'FT8',
    #     'RST_SENT': '',
    #     'RST_RCVD': '',
    #     'QSO_DATE': '20210329',
    #     'TIME_ON': '183213',
    #     'QSO_DATE_OFF': '20210329',
    #     'TIME_OFF': '183213',
    #     'BAND': '20M',
    #     'FREQ': '14.074754',
    #     'STATION_CALLSIGN': 'K6GTE',
    #     'MY_GRIDSQUARE': 'DM13AT',
    #     'CONTEST_ID': 'ARRL-FIELD-DAY',
    #     'SRX_STRING': '1D UT',
    #     'CLASS': '1D',
    #     'ARRL_SECT': 'UT'
    # }
    # FlDigi
    # {
    #     'CALL': 'K5TUS', 
    #     'MODE': 'RTTY', 
    #     'FREQ': '14.068415', 
    #     'BAND': '20M', 
    #     'QSO_DATE': '20250103', 
    #     'TIME_ON': '2359', 
    #     'QSO_DATE_OFF': '20250103', 
    #     'TIME_OFF': '2359', 
    #     'NAME': '', 
    #     'QTH': '', 
    #     'STATE': 'ORG', 
    #     'VE_PROV': '', 
    #     'COUNTRY': 'USA', 
    #     'RST_SENT': '599', 
    #     'RST_RCVD': '599', 
    #     'TX_PWR': '0', 
    #     'CNTY': '', 
    #     'DXCC': '', 
    #     'CQZ': '5', 
    #     'IOTA': '', 
    #     'CONT': '', 
    #     'ITUZ': '', 
    #     'GRIDSQUARE': '', 
    #     'QSLRDATE': '', 
    #     'QSLSDATE': '', 
    #     'EQSLRDATE': '', 
    #     'EQSLSDATE': '', 
    #     'LOTWRDATE': '', 
    #     'LOTWSDATE': '', 
    #     'QSL_VIA': '', 
    #     'NOTES': '', 
    #     'SRX': '', 
    #     'STX': '000', 
    #     'SRX_STRING': '', 
    #     'STX_STRING': 'CA', 


    #     'SRX': '666', 
    #     'STX': '000', 
    #     'SRX_STRING': '', 
    #     'STX_STRING': 'CA',

    #     'SRX': '004', 'STX': '000', 'SRX_STRING': '', 'STX_STRING': '#',

    #     'CLASS': '', 
    #     'ARRL_SECT': '', 
    #     'OPERATOR': 'K6GTE', 
    #     'STATION_CALLSIGN': 'K6GTE', 
    #     'MY_GRIDSQUARE': 'DM13AT', 
    #     'MY_CITY': 'ANAHEIM, CA', 
    #     'CHECK': '', 
    #     'AGE': '', 
    #     'TEN_TEN': '', 
    #     'CWSS_PREC': '', 
    #     'CWSS_SECTION': '', 
    #     'CWSS_SERNO': '', 
    #     'CWSS_CHK': ''
    # }

    # """

    # logger.debug(f"{the_packet=}")
    # if ALTEREGO is not None:
    #     ALTEREGO.callsign.setText(the_packet.get("CALL"))
    #     ALTEREGO.contact["Call"] = the_packet.get("CALL", "")
    #     ALTEREGO.contact["SNT"] = the_packet.get("RST_SENT", "599")
    #     ALTEREGO.contact["RCV"] = the_packet.get("RST_RCVD", "599")

    #     sent_string = the_packet.get("STX_STRING", "")
    #     if sent_string != "":
    #         ALTEREGO.contact["SentNr"] = sent_string
    #         ALTEREGO.other_1.setText(str(sent_string))
    #     else:
    #         ALTEREGO.contact["SentNr"] = the_packet.get("STX", "000")
    #         ALTEREGO.other_1.setText(str(the_packet.get("STX", "000")))

    #     rx_string = the_packet.get("SRX_STRING", "")
    #     if rx_string != "":
    #         ALTEREGO.contact["NR"] = rx_string
    #         ALTEREGO.other_2.setText(str(rx_string))
    #     else:
    #         ALTEREGO.contact["NR"] = the_packet.get("SRX", "000")
    #         ALTEREGO.other_2.setText(str(the_packet.get("SRX", "000")))

    #     ALTEREGO.contact["Mode"] = the_packet.get("MODE", "ERR")
    #     ALTEREGO.contact["Freq"] = round(float(the_packet.get("FREQ", "0.0")) * 1000, 2)
    #     ALTEREGO.contact["QSXFreq"] = round(
    #         float(the_packet.get("FREQ", "0.0")) * 1000, 2
    #     )
    #     ALTEREGO.contact["Band"] = get_logged_band(
    #         str(int(float(the_packet.get("FREQ", "0.0")) * 1000000))
    #     )
    #     logger.debug(f"{ALTEREGO.contact=}")

    #     ALTEREGO.save_contact()



""" """

import logging
import platform
from datetime import datetime
from datetime import timedelta
from datetime import timezone


from pathlib import Path

from PyQt6 import QtWidgets

from not1mm.lib.plugin_common import gen_adif, imp_adif, get_points

from not1mm.lib.ham_utility import calculate_wpx_prefix
from not1mm.lib.version import __version__

logger = logging.getLogger(__name__)

EXCHANGE_HINT = "#"


name = "ES OPEN"
cabrillo_name = "ES-OPEN"
mode = "BOTH"  # CW SSB BOTH RTTY

columns = [
    "YYYY-MM-DD HH:MM:SS",
    "Call",
    "Freq",
    "Mode",
    "Snt",
    "Rcv",
    "SentNr",
    "RcvNr",
    "PTS",
]

advance_on_space = [True, True, True, True, True]

# 5 Contest specific dupe check.
dupe_type = 5


def specific_contest_check_dupe(self, call):
    """"""
    # get mode from radio state
    mode = self.radio_state.get("mode", "")
    """Dupe checking specific to just this contest."""
    # constant to split the contest - correct ES Open Contest length is 4 hours
    contest_length_in_minutes = 240
    split_contest_by_minutes = 60

    period_count = int(contest_length_in_minutes / split_contest_by_minutes)

    # think about generic solution by splitting the contest to n different periods
    start_date_init = self.contest_settings.get("StartDate", "")
    start_date_init_date = datetime.strptime(start_date_init, "%Y-%m-%d %H:%M:%S")

    # Create time periods dynamically based on period count
    time_periods = []
    for i in range(period_count):
        minutes_to_add = split_contest_by_minutes * (i + 1)
        time_period = start_date_init_date + timedelta(minutes=minutes_to_add)
        time_periods.append(time_period)

    # Assign to variables for backwards compatibility
    time_period_1 = time_periods[0] if len(time_periods) > 0 else None
    time_period_2 = time_periods[1] if len(time_periods) > 1 else None
    time_period_3 = time_periods[2] if len(time_periods) > 2 else None
    time_period_4 = time_periods[3] if len(time_periods) > 3 else None

    # get current time in UTC
    iso_current_time = datetime.now(timezone.utc)
    current_time = iso_current_time.replace(tzinfo=None)

    if (
        time_period_1 is not None
        and current_time < time_period_1
        and current_time >= start_date_init_date
    ):

        return self.database.check_dupe_on_period_mode(
            call,
            self.contact.get("Band", ""),
            mode,
            start_date_init_date,
            time_period_1.strftime("%Y-%m-%d %H:%M:%S"),
        )

    if (
        time_period_1 is not None
        and time_period_2 is not None
        and current_time < time_period_2
        and current_time >= time_period_1
    ):

        return self.database.check_dupe_on_period_mode(
            call,
            self.contact.get("Band", ""),
            mode,
            time_period_1.strftime("%Y-%m-%d %H:%M:%S"),
            time_period_2.strftime("%Y-%m-%d %H:%M:%S"),
        )

    if (
        time_period_2 is not None
        and time_period_3 is not None
        and current_time < time_period_3
        and current_time >= time_period_2
    ):

        return self.database.check_dupe_on_period_mode(
            call,
            self.contact.get("Band", ""),
            mode,
            time_period_2.strftime("%Y-%m-%d %H:%M:%S"),
            time_period_3.strftime("%Y-%m-%d %H:%M:%S"),
        )

    if (
        time_period_3 is not None
        and time_period_4 is not None
        and current_time < time_period_4
        and current_time >= time_period_3
    ):

        return self.database.check_dupe_on_period_mode(
            call,
            self.contact.get("Band", ""),
            mode,
            time_period_3.strftime("%Y-%m-%d %H:%M:%S"),
            time_period_4.strftime("%Y-%m-%d %H:%M:%S"),
        )
    # just for band and mode if outside of time period
    return self.database.check_dupe_on_band_mode(
        call, self.contact.get("Band", ""), mode
    )


def init_contest(self):
    """setup plugin"""
    set_tab_next(self)
    set_tab_prev(self)
    interface(self)
    self.next_field = self.other_2


def interface(self):
    """Setup user interface"""
    self.field1.show()
    self.field2.show()
    self.field3.show()
    self.field4.show()
    self.other_label.setText("Sent")
    self.field3.setAccessibleName("Sent")
    self.exch_label.setText("SN")
    self.field4.setAccessibleName("SN")


def reset_label(self):
    """reset label after field cleared"""


def set_tab_next(self):
    """Set TAB Advances"""
    self.tab_next = {
        self.callsign: self.other_1,
        self.sent: self.other_1,
        self.receive: self.other_1,
        self.other_1: self.other_2,
        self.other_2: self.callsign,
    }


def set_tab_prev(self):
    """Set TAB Advances"""
    self.tab_prev = {
        self.callsign: self.other_2,
        self.sent: self.callsign,
        self.receive: self.callsign,
        self.other_1: self.callsign,
        self.other_2: self.other_1,
    }


def set_contact_vars(self):
    """Contest Specific"""
    self.contact["SNT"] = self.sent.text()
    self.contact["RCV"] = self.receive.text()
    self.contact["SentNr"] = self.other_1.text().upper()
    self.contact["NR"] = self.other_2.text().upper()

    self.contact["IsMultiplier1"] = 0
    self.contact["IsMultiplier2"] = 0


def predupe(self):
    """called after callsign entered"""


def prefill(self):
    """Fill SentNR"""
    sent_sxchange_setting = self.contest_settings.get("SentExchange", "")
    if sent_sxchange_setting.strip() == "#":
        # result = self.database.get_serial()
        # serial_nr = str(result.get("serial_nr", "1")).zfill(3)
        serial_nr = str(self.current_sn).zfill(3)
        if serial_nr == "None":
            serial_nr = "001"
        if len(self.other_1.text()) == 0:
            self.other_1.setText(serial_nr)
    else:
        self.other_1.setText(sent_sxchange_setting)


def points(self):
    """ """
    if self.contact_is_dupe > 0:
        return 0

    _mode = self.contact.get("Mode", "")
    if _mode in "SSB, USB, LSB, FM, AM":
        return 1
    if _mode in "CW":
        return 2

    return 0


def show_mults(self, rtc=None):
    """Return display string for mults"""
    our_prefix = calculate_wpx_prefix(self.station.get("Call", ""))
    query = f"SELECT count(DISTINCT(substr(WPXPrefix,3,1) || ':' || Band || ':' || Mode)) as mults from DXLOG where ContestNR = {self.pref.get('contest', '1')} AND CountryPrefix = 'ES' AND WPXPrefix != '{our_prefix}';"
    result = self.database.exec_sql(query)
    if result:
        mult_count = result.get("mults", 0)
        return mult_count
    return 0


def show_qso(self):
    """Return qso count"""
    result = self.database.fetch_qso_count()
    if result:
        return int(result.get("qsos", 0))
    return 0


def calc_score(self):
    """Return calculated score"""
    result = self.database.fetch_points()
    if result is not None:
        score = result.get("Points", "0")
        if score is None:
            score = "0"
        contest_points = int(score)
        mults = show_mults(self)
        return contest_points * (mults + 1)
    return 0


def recalculate_mults(self):
    """Recalculates multipliers after change in logged qso."""


def adif(self):
    """Call the generate ADIF function"""
    gen_adif(self, cabrillo_name, "ES OPEN")


def output_cabrillo_line(line_to_output, ending, file_descriptor, file_encoding):
    """"""
    print(
        line_to_output.encode(file_encoding, errors="ignore").decode(),
        end=ending,
        file=file_descriptor,
    )


def cabrillo(self, file_encoding):
    """Generates Cabrillo file. Maybe."""
    # https://www.cqwpx.com/cabrillo.htm
    logger.debug("******Cabrillo*****")
    logger.debug("Station: %s", f"{self.station}")
    logger.debug("Contest: %s", f"{self.contest_settings}")
    now = datetime.now()
    date_time = now.strftime("%Y-%m-%d_%H-%M-%S")
    filename = (
        str(Path.home())
        + "/"
        + f"{self.station.get('Call', '').upper().replace('/','-')}_{cabrillo_name}_{date_time}.log"
    )
    logger.debug("%s", filename)
    log = self.database.fetch_all_contacts_asc()
    try:
        with open(filename, "w", encoding=file_encoding, newline="") as file_descriptor:
            output_cabrillo_line(
                "START-OF-LOG: 3.0",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CREATED-BY: Not1MM v{__version__}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CONTEST: {cabrillo_name}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.station.get("Club", ""):
                output_cabrillo_line(
                    f"CLUB: {self.station.get('Club', '').upper()}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"CALLSIGN: {self.station.get('Call','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"LOCATION: {self.station.get('ARRLSection', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-OPERATOR: {self.contest_settings.get('OperatorCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-ASSISTED: {self.contest_settings.get('AssistedCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-BAND: {self.contest_settings.get('BandCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            mode = self.contest_settings.get("ModeCategory", "")
            if mode in ["SSB+CW", "SSB+CW+DIGITAL"]:
                mode = "MIXED"
            output_cabrillo_line(
                f"CATEGORY-MODE: {mode}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-TRANSMITTER: {self.contest_settings.get('TransmitterCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.contest_settings.get("OverlayCategory", "") != "N/A":
                output_cabrillo_line(
                    f"CATEGORY-OVERLAY: {self.contest_settings.get('OverlayCategory','')}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"GRID-LOCATOR: {self.station.get('GridSquare','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-POWER: {self.contest_settings.get('PowerCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )

            output_cabrillo_line(
                f"CLAIMED-SCORE: {calc_score(self)}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            ops = ""
            list_of_ops = self.database.get_ops()
            for op in list_of_ops:
                ops += f"{op.get('Operator', '')}, "
            if self.station.get("Call", "") not in ops:
                ops += f"@{self.station.get('Call','')}"
            else:
                ops = ops.rstrip(", ")
            output_cabrillo_line(
                f"OPERATORS: {ops}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"NAME: {self.station.get('Name', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('Street1', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-CITY: {self.station.get('City', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-STATE-PROVINCE: {self.station.get('State', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-POSTALCODE: {self.station.get('Zip', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-COUNTRY: {self.station.get('Country', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"EMAIL: {self.station.get('Email', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            for contact in log:
                the_date_and_time = contact.get("TS", "")
                themode = contact.get("Mode", "")
                if themode in ("CW-U", "CW-L", "CW-R", "CWR"):
                    themode = "CW"
                if themode == "LSB" or themode == "USB":
                    themode = "PH"
                frequency = str(round(contact.get("Freq", "0"))).rjust(5)

                loggeddate = the_date_and_time[:10]
                loggedtime = the_date_and_time[11:13] + the_date_and_time[14:16]
                output_cabrillo_line(
                    f"QSO: {frequency} {themode} {loggeddate} {loggedtime} "
                    f"{contact.get('StationPrefix', '').ljust(13)} "
                    f"{str(contact.get('SNT', '')).ljust(3)} "
                    f"{str(contact.get('SentNr', '')).ljust(6)} "
                    f"{contact.get('Call', '').ljust(13)} "
                    f"{str(contact.get('RCV', '')).ljust(3)} "
                    f"{str(contact.get('NR', '')).ljust(6)}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line("END-OF-LOG:", "\r\n", file_descriptor, file_encoding)
        self.show_message_box(f"Cabrillo saved to: {filename}")
    except IOError as exception:
        logger.critical("cabrillo: IO error: %s, writing to %s", exception, filename)
        self.show_message_box(f"Error saving Cabrillo: {exception} {filename}")
        return


# def populate_history_info_line(self):
#     result = self.database.fetch_call_history(self.callsign.text())
#     if result:
#         self.history_info.setText(
#             f"{result.get('Call', '')}, {result.get('Exch1', '')}, {result.get('UserText','...')}"
#         )
#     else:
#         self.history_info.setText("")


# def check_call_history(self):
#     """"""
#     result = self.database.fetch_call_history(self.callsign.text())
#     print(f"{result=}")
#     if result:
#         self.history_info.setText(f"{result.get('UserText','')}")
#         if self.other_2.text() == "":
#             self.other_2.setText(f"{result.get('Exch1', '')}")


def process_esm(self, new_focused_widget=None, with_enter=False):
    """ESM State Machine"""

    # self.pref["run_state"]

    # -----===== Assigned F-Keys =====-----
    # self.esm_dict["CQ"]
    # self.esm_dict["EXCH"]
    # self.esm_dict["QRZ"]
    # self.esm_dict["AGN"]
    # self.esm_dict["HISCALL"]
    # self.esm_dict["MYCALL"]
    # self.esm_dict["QSOB4"]

    # ----==== text fields ====----
    # self.callsign
    # self.sent
    # self.receive
    # self.other_1
    # self.other_2

    if new_focused_widget is not None:
        self.current_widget = self.inputs_dict.get(new_focused_widget)

    # print(f"checking esm {self.current_widget=} {with_enter=} {self.pref.get("run_state")=}")

    for a_button in [
        self.esm_dict["CQ"],
        self.esm_dict["EXCH"],
        self.esm_dict["QRZ"],
        self.esm_dict["AGN"],
        self.esm_dict["HISCALL"],
        self.esm_dict["MYCALL"],
        self.esm_dict["QSOB4"],
    ]:
        if a_button is not None:
            self.restore_button_color(a_button)

    buttons_to_send = []

    if self.pref.get("run_state"):
        if self.current_widget == "callsign":
            if len(self.callsign.text()) < 3:
                self.make_button_green(self.esm_dict["CQ"])
                buttons_to_send.append(self.esm_dict["CQ"])
            elif len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["HISCALL"])
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["HISCALL"])
                buttons_to_send.append(self.esm_dict["EXCH"])

        elif self.current_widget == "other_2":
            if self.other_2.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["QRZ"])
                buttons_to_send.append(self.esm_dict["QRZ"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    self.process_function_key(button)
    else:
        if self.current_widget == "callsign":
            if len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["MYCALL"])
                buttons_to_send.append(self.esm_dict["MYCALL"])

        elif self.current_widget == "other_2":
            if self.other_2.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["EXCH"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    self.process_function_key(button)


# def get_mults(self):
#     """Get mults for RTC XML"""
#     mults = {}
#     mults["country"], mults["state"] = show_mults(self, rtc=True)
#     return mults


# def just_points(self):
#     """Get points for RTC XML"""
#     return get_points(self)



"""Dutch PACC Contest"""

"""
6. Exchange:
6.1 Non-Dutch (Non-PA) stations: signal report + QSO number, starting with 001.
6.1.1 Non-Dutch (Non-PA) MOAB (Multi Op, All Bands) stations can use separate serial numbers for each band if using
more than one transmitter at a time. If only one transmitter at a time is used, chronological serial numbers need to be used,
but skipped numbers are allowed (e.g. as a result of blocking numbers for the multiplier station).
6.2 Dutch (PA) stations: signal report + province code (two letters).
There are 12 provinces: DR, FL, FR, GD, GR, LB, NB, NH, OV, UT, ZH and ZL
7. QSO Points
7.1 Each valid contact with another participating contest station is worth one (1) point. For foreign stations: only contacts
with Dutch (PA) stations count. In MIXED mode (CW/SSB) a valid qso counts once per mode per band. For PA stations:
cross contest contacts (e.g. if the PACC coincides with the RSGB 160m contest) are considered valid as long as a serial
number is received. No cross mode or cross band contacts are allowed.
7.2 SWL stations count each individual heard PACC contest station (both Dutch (PA) and non-Dutch (Non-PA)) as 1 point
provided both the calls and the exchange number of the heard station (i.e. RST + number or RST + Province) is logged of
the QSO of concern. Each call counts only once per band, per mode and one particular contest station may not be logged
as counter station more than 10 times per band.
7.2.1 All SWL QSO’s should be logged during the contest.
8. Dupes
8.1 Dupes are contacts made with the same station on the same band. If the first contact between stations is valid, dupes
have 0 points value. If the first contact is not valid, the second (dupe) contact is accepted.
8.2 DO NOT DELETE DUPES! Dupe contacts are not penalized; one does not have to mark them in the Cabrillo log
submission.
Moreover, entrants are strongly recommended to leave DUPES in the log file.
9. Multipliers
9.1 Non-Dutch (Non-PA) stations (incl. SWL stations): a multiplier of one (1) for each different Dutch (PA) province
contacted on each band on each mode.
The Dutch (PA) province abbreviations are:
DR = Drenthe NB = Noord-Brabant
FL = Flevoland NH = Noord-Holland
FR = Friesland OV = Overijssel
GD = Gelderland UT = Utrecht
GR = Groningen ZH = Zuid-Holland
LB = Limburg ZL = Zeeland
9.2 Dutch (PA) stations (incl. SWL stations):
• A multiplier of one (1) for each different DXCC entity (ARRL list) on each band, each mode including PA, with the
following exceptions:
• A multiplier of one (1) for each different call area of Asiatic Russia (UA8/9/0), Chile (CE), Japan (JA), Argentina (LU),
Brazil (PY), Canada (VE), USA, Australia (VK), South Africa (ZS) and New-Zealand (ZL) on each band, each mode.
Note 1: this means that Asiatic Russia has 3 call areas/multipliers: UA8, UA9 and UA0. Canadian call areas are interpreted
as geographical districts, which means VE1, VO1, VY1, VE2, VO2, VY2, VY0 are all separate multipliers, but e.g. VE2,
CG2, XK2 are all the same multiplier (VE2).
Note 2: A reciprocal call sign in one of the countries which have call areas as multipliers, but without a call area designation,
is counted as the zero-call area, e.g. LU/G3XYZ gives multiplier LU0. Reciprocal calls in W, JA, VE and UA need to indicate
their call area, otherwise the call should be regarded as invalid, e.g. W3/DL8ABC is valid, W/DL8ABC is invalid. (This
means that, theoretically, PY0 is a separate multiplier besides PY0F, PY0S and PY0T).
Note 3: Call areas are taken as-is from the call sign, even if it is known that the station is located in a different area,
UNLESS the station explicitly indicates it. E.g. K5ZD is multiplier W5, but K5ZD/1 would give multiplier W1.
Note 4: Special call signs will count for the call area to which they geographically belong (e.g. UE150SBM = UA0).
"""

"""
Test calls

PA2WRD             PA3ADU             PA3AKP             PA3ARM
PA3BGQ             PA3BQP             PA3BQS             PA3CDD
PA3CSG             PA3CXB              PA3DB             PA3DNA
PA3DRL             PA3DSB             PA3DTR             PA3DUU
PA3DVA             PA3DWJ             PA3EEG             PA3EKM
PA3ELQ              PA3EM             PA3EMN             PA3EOU
PA3EOZ             PA3EPO             PA3EVY             PA3EWG
PA3FUJ             PA3FZV             PA3GDD             PA3GEO
PA3GMM  
"""
# pylint: disable=invalid-name, unused-argument, unused-variable, c-extension-no-member, unused-import

import datetime
import logging

from pathlib import Path

from PyQt6 import QtWidgets

from not1mm.lib.plugin_common import gen_adif, imp_adif, get_points, online_score_xml
from not1mm.lib.version import __version__

logger = logging.getLogger(__name__)

EXCHANGE_HINT = "# or Province"

name = "PACC"
cabrillo_name = "PACC"
mode = "BOTH"  # CW SSB BOTH RTTY

columns = [
    "YYYY-MM-DD HH:MM:SS",
    "Call",
    "Freq",
    "Snt",
    "Rcv",
    "SentNr",
    "RcvNr",
    "PTS",
]

advance_on_space = [True, True, True, True, True]

# 1 once per contest, 2 work each band, 3 each band/mode, 4 no dupe checking
dupe_type = 3


PROV_SEC = [
    "DR",
    "FL",
    "FR",
    "GD",
    "GR",
    "LB",
    "NB",
    "NH",
    "OV",
    "UT",
    "ZH",
    "ZL",
]


def init_contest(self):
    """setup plugin"""
    set_tab_next(self)
    set_tab_prev(self)
    interface(self)
    self.next_field = self.other_2


def interface(self):
    """Setup user interface"""
    self.field1.show()
    self.field2.show()
    self.field3.show()
    self.field4.show()
    self.snt_label.setText("SNT")
    self.field1.setAccessibleName("RST Sent")
    self.other_label.setText("SentNR")
    self.field3.setAccessibleName("Sent Number")
    self.exch_label.setText("Prov or SN")
    self.field4.setAccessibleName("Province code or Serial Number")


def reset_label(self):
    """reset label after field cleared"""


def set_tab_next(self):
    """Set TAB Advances"""
    self.tab_next = {
        self.callsign: self.sent,
        self.sent: self.receive,
        self.receive: self.other_1,
        self.other_1: self.other_2,
        self.other_2: self.callsign,
    }


def set_tab_prev(self):
    """Set TAB Advances"""
    self.tab_prev = {
        self.callsign: self.other_2,
        self.sent: self.callsign,
        self.receive: self.sent,
        self.other_1: self.receive,
        self.other_2: self.other_1,
    }


def validate(self):
    """doc"""
    return True


def set_contact_vars(self):
    """Contest Specific"""
    self.contact["SNT"] = self.sent.text()
    self.contact["RCV"] = self.receive.text()
    self.contact["NR"] = self.other_2.text().upper()
    self.contact["SentNr"] = self.other_1.text()


def predupe(self):
    """called after callsign entered"""


def prefill(self):
    """Fill sentnr"""
    # result = self.database.get_serial()
    # serial_nr = str(result.get("serial_nr", "1")).zfill(3)
    serial_nr = str(self.current_sn).zfill(3)
    if serial_nr == "None":
        serial_nr = "001"

    exchange = self.contest_settings.get("SentExchange", "").replace("#", serial_nr)
    if len(self.other_1.text()) == 0:
        self.other_1.setText(exchange)


def points(self):
    """Calc point"""

    # Each valid contact with another participating contest station is worth one (1) point. For foreign stations: only contacts
    # with Dutch (PA) stations count. In MIXED mode (CW/SSB) a valid qso counts once per mode per band. For PA stations:
    # cross contest contacts (e.g. if the PACC coincides with the RSGB 160m contest) are considered valid as a long as a serial
    # number is received. No cross mode or cross band contacts are allowed.

    if self.contact_is_dupe > 0:
        return 0

    # Am I a dutch station.
    im_dutch = False
    theyre_dutch = False

    my_exchange = str(self.contest_settings.get("SentExchange", 0))
    if my_exchange.isalpha():
        im_dutch = True

    their_exchange = self.other_2.text().upper()
    if their_exchange.isalpha():
        theyre_dutch = True

    if im_dutch:
        return 1
    else:
        if theyre_dutch:
            return 1

    return 0


def im_dutch(self):
    my_exchange = str(self.contest_settings.get("SentExchange", 0))
    if my_exchange.isalpha():
        return True
    return False


def show_mults(self):
    """Return display string for mults"""

    sql = ""
    if im_dutch(self) is False:
        sql = f"select count(DISTINCT(NR || ':' || Band || ':' || Mode)) as mult_count from dxlog where ContestNR = {self.database.current_contest} and NR in ('DR','FL','FR','GD','GR','LB','NB','NH','OV','UT','ZH','ZL');"
    else:
        sql = f"select count(DISTINCT(CountryPrefix || ':' || Band || ':' || Mode)) as mult_count from dxlog where ContestNR = {self.database.current_contest};"

    result = self.database.exec_sql(sql)
    if result:
        return result.get("mult_count", 0)
    return 0


def show_qso(self):
    """Return qso count"""
    result = self.database.fetch_qso_count()
    if result:
        return int(result.get("qsos", 0))
    return 0


def calc_score(self):
    """Return calculated score"""
    mults = show_mults(self)
    result = self.database.fetch_points()
    if result is not None:
        score = result.get("Points", "0")
        if score is None:
            score = "0"
        if int(mults) > 0:
            contest_points = int(score) * (int(mults) + 1)
            return contest_points
        contest_points = int(score)
        return contest_points
    return 0


def adif(self):
    """Call the generate ADIF function"""
    gen_adif(self, cabrillo_name, "PACC")


def output_cabrillo_line(line_to_output, ending, file_descriptor, file_encoding):
    """"""
    print(
        line_to_output.encode(file_encoding, errors="ignore").decode(),
        end=ending,
        file=file_descriptor,
    )


def cabrillo(self, file_encoding):
    """Generates Cabrillo file. Maybe."""
    # https://www.cqwpx.com/cabrillo.htm
    logger.debug("******Cabrillo*****")
    logger.debug("Station: %s", f"{self.station}")
    logger.debug("Contest: %s", f"{self.contest_settings}")
    now = datetime.datetime.now()
    date_time = now.strftime("%Y-%m-%d_%H-%M-%S")
    filename = (
        str(Path.home())
        + "/"
        + f"{self.station.get('Call', '').upper().replace('/','-')}_{cabrillo_name}_{date_time}.log"
    )
    logger.debug("%s", filename)
    log = self.database.fetch_all_contacts_asc()
    try:
        with open(filename, "w", encoding=file_encoding, newline="") as file_descriptor:
            output_cabrillo_line(
                "START-OF-LOG: 3.0",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CREATED-BY: Not1MM v{__version__}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CONTEST: {cabrillo_name}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.station.get("Club", ""):
                output_cabrillo_line(
                    f"CLUB: {self.station.get('Club', '').upper()}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"CALLSIGN: {self.station.get('Call','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"LOCATION: {self.station.get('ARRLSection', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-OPERATOR: {self.contest_settings.get('OperatorCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-ASSISTED: {self.contest_settings.get('AssistedCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-BAND: {self.contest_settings.get('BandCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            mode = self.contest_settings.get("ModeCategory", "")
            if mode in ["SSB+CW", "SSB+CW+DIGITAL"]:
                mode = "MIXED"
            output_cabrillo_line(
                f"CATEGORY-MODE: {mode}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-TRANSMITTER: {self.contest_settings.get('TransmitterCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.contest_settings.get("OverlayCategory", "") != "N/A":
                output_cabrillo_line(
                    f"CATEGORY-OVERLAY: {self.contest_settings.get('OverlayCategory','')}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"GRID-LOCATOR: {self.station.get('GridSquare','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-POWER: {self.contest_settings.get('PowerCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )

            output_cabrillo_line(
                f"CLAIMED-SCORE: {calc_score(self)}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            ops = ""
            list_of_ops = self.database.get_ops()
            for op in list_of_ops:
                ops += f"{op.get('Operator', '')}, "
            if self.station.get("Call", "") not in ops:
                ops += f"@{self.station.get('Call','')}"
            else:
                ops = ops.rstrip(", ")
            output_cabrillo_line(
                f"OPERATORS: {ops}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"NAME: {self.station.get('Name', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('Street1', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-CITY: {self.station.get('City', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-STATE-PROVINCE: {self.station.get('State', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-POSTALCODE: {self.station.get('Zip', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-COUNTRY: {self.station.get('Country', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"EMAIL: {self.station.get('Email', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            for contact in log:
                the_date_and_time = contact.get("TS", "")
                themode = contact.get("Mode", "")
                if themode in ("CW-U", "CW-L", "CW-R", "CWR"):
                    themode = "CW"
                if themode in ("LSB", "USB", "AM", "FM"):
                    themode = "PH"
                frequency = str(round(contact.get("Freq", "0"))).rjust(5)

                loggeddate = the_date_and_time[:10]
                loggedtime = the_date_and_time[11:13] + the_date_and_time[14:16]
                output_cabrillo_line(
                    f"QSO: {frequency} {themode} {loggeddate} {loggedtime} "
                    f"{contact.get('StationPrefix', '').ljust(13)} "
                    f"{str(contact.get('SNT', '')).ljust(3)} "
                    f"{str(contact.get('SentNr', '')).ljust(6)} "
                    f"{contact.get('Call', '').ljust(13)} "
                    f"{str(contact.get('RCV', '')).ljust(3)} "
                    f"{str(contact.get('NR', '')).ljust(6)}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line("END-OF-LOG:", "\r\n", file_descriptor, file_encoding)
        self.show_message_box(f"Cabrillo saved to: {filename}")
    except IOError as exception:
        logger.critical("cabrillo: IO error: %s, writing to %s", exception, filename)
        self.show_message_box(f"Error saving Cabrillo: {exception} {filename}")
        return


def recalculate_mults(self):
    """Recalculates multipliers after change in logged qso."""


def process_esm(self, new_focused_widget=None, with_enter=False):
    """ESM State Machine"""

    # self.pref["run_state"]

    # -----===== Assigned F-Keys =====-----
    # self.esm_dict["CQ"]
    # self.esm_dict["EXCH"]
    # self.esm_dict["QRZ"]
    # self.esm_dict["AGN"]
    # self.esm_dict["HISCALL"]
    # self.esm_dict["MYCALL"]
    # self.esm_dict["QSOB4"]

    # ----==== text fields ====----
    # self.callsign
    # self.sent
    # self.receive
    # self.other_1
    # self.other_2

    if new_focused_widget is not None:
        self.current_widget = self.inputs_dict.get(new_focused_widget)

    for a_button in [
        self.esm_dict["CQ"],
        self.esm_dict["EXCH"],
        self.esm_dict["QRZ"],
        self.esm_dict["AGN"],
        self.esm_dict["HISCALL"],
        self.esm_dict["MYCALL"],
        self.esm_dict["QSOB4"],
    ]:
        if a_button is not None:
            self.restore_button_color(a_button)

    buttons_to_send = []

    if self.pref.get("run_state"):
        if self.current_widget == "callsign":
            if len(self.callsign.text()) < 3:
                self.make_button_green(self.esm_dict["CQ"])
                buttons_to_send.append(self.esm_dict["CQ"])
            elif len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["HISCALL"])
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["HISCALL"])
                buttons_to_send.append(self.esm_dict["EXCH"])

        elif self.current_widget in ["other_1", "other_2"]:
            if self.other_1.text() == "" or self.other_2.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["QRZ"])
                buttons_to_send.append(self.esm_dict["QRZ"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    self.process_function_key(button)
    else:
        if self.current_widget == "callsign":
            if len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["MYCALL"])
                buttons_to_send.append(self.esm_dict["MYCALL"])

        elif self.current_widget in ["other_1", "other_2"]:
            if self.other_1.text() == "" or self.other_2.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["EXCH"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    self.process_function_key(button)


def get_mults(self):
    """"""

    mults = {}
    mults["state"] = show_mults(self)
    return mults


def just_points(self):
    """"""
    result = self.database.fetch_points()
    if result is not None:
        score = result.get("Points", "0")
        if score is None:
            score = "0"
        return int(score)
    return 0


def populate_history_info_line(self):
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        self.history_info.setText(
            f"{result.get('Call', '')}, {result.get('Sect', '')}, {result.get('UserText','...')}"
        )
    else:
        self.history_info.setText("")


def check_call_history(self):
    """"""
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        self.history_info.setText(f"{result.get('UserText','')}")
        # if self.other_1.text() == "":
        #     self.other_1.setText(f"{result.get('Exch1', '')}")
        if self.other_2.text() == "":
            self.other_2.setText(f"{result.get('Sect', '')}")



"""SSA Manadstest CW plugin"""

# SSA Manadstest - Swedish monthly HF contest, CW mode.
#   Exchange:   RST + serial number + 6-character Maidenhead locator
#   Example:    599 01 JP75XX
#   Multipliers: unique 4-character locator prefixes per band
#   Score:      total QSO points * total multipliers
#   QSO points: 2 per valid QSO

import datetime
import logging

from pathlib import Path
from not1mm.lib.plugin_common import gen_adif, imp_adif, get_points
from not1mm.lib.version import __version__

logger = logging.getLogger(__name__)

VALID_LOCATORS = {
    "JO57",
    "JO58",
    "JO59",
    "JO65",
    "JO66",
    "JO67",
    "JO68",
    "JO69",
    "JO75",
    "JO76",
    "JO77",
    "JO78",
    "JO79",
    "JO86",
    "JO87",
    "JO88",
    "JO89",
    "JO96",
    "JO97",
    "JO98",
    "JO99",
    "JP53",
    "JP60",
    "JP61",
    "JP62",
    "JP63",
    "JP64",
    "JP70",
    "JP71",
    "JP72",
    "JP73",
    "JP74",
    "JP75",
    "JP76",
    "JP80",
    "JP81",
    "JP82",
    "JP83",
    "JP84",
    "JP85",
    "JP86",
    "JP87",
    "JP88",
    "JP90",
    "JP92",
    "JP93",
    "JP94",
    "JP95",
    "JP96",
    "JP97",
    "JP98",
    "KP03",
    "KP04",
    "KP05",
    "KP06",
    "KP07",
    "KP08",
    "KP09",
    "KP15",
    "KP16",
    "KP17",
    "KP18",
    "KP25",
    "KP26",
}

EXCHANGE_HINT = "# LocXXXX"

name = "SSA MT CW"
cabrillo_name = "SSA-MT-CW"
mode = "CW"
columns = [
    "YYYY-MM-DD HH:MM:SS",
    "Call",
    "Freq (KHz)",
    "Snt",
    "Rcv",
    "SentNr",
    "RcvNr",
    "Exchange1",
    "PTS",
]

advance_on_space = [True, True, True, True, True]

# 1 once per contest, 2 once per band, 3 once per band/mode, 4 no dupe checking
dupe_type = 2


def init_contest(self):
    """Setup plugin"""
    set_tab_next(self)
    set_tab_prev(self)
    interface(self)
    self.next_field = self.other_2
    self._exch_sent = False


def interface(self):
    """Setup user interface"""
    self.field1.show()
    self.field2.show()
    self.field3.show()
    self.field4.show()
    self.sent.setReadOnly(False)
    self.snt_label.setText("Sent")
    self.rcv_label.setText("Received")
    self.field1.setAccessibleName("Signal Report Sent")
    self.other_label.setText("Nr")
    self.field3.setAccessibleName("Received Serial Number")
    self.exch_label.setText("Locator")
    self.field4.setAccessibleName("Received Locator")


def reset_label(self):
    """Reset label after field cleared"""


def set_tab_next(self):
    """Set TAB advance order"""
    self.tab_next = {
        self.callsign: self.sent,
        self.sent: self.receive,
        self.receive: self.other_1,
        self.other_1: self.other_2,
        self.other_2: self.callsign,
    }


def set_tab_prev(self):
    """Set reverse TAB order"""
    self.tab_prev = {
        self.callsign: self.other_2,
        self.sent: self.callsign,
        self.receive: self.sent,
        self.other_1: self.receive,
        self.other_2: self.other_1,
    }


def set_contact_vars(self):
    """Map UI fields to contact record"""
    self.contact["SNT"] = self.sent.text()
    self.contact["RCV"] = self.receive.text()
    self.contact["SentNr"] = self.current_sn
    self.contact["GridSquare"] = self.contest_settings.get("SentExchange", "")
    self.contact["NR"] = self.other_1.text()
    self.contact["Exchange1"] = self.other_2.text().upper()


def predupe(self):
    """Pre-fill locator from call history on each callsign keystroke"""
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        name = result.get("Name", "") or ""
        loc = result.get("Loc1", "") or ""
        info = ", ".join(x for x in [result.get("Call", ""), name, loc] if x)
        self.history_info.setText(info)
        self.other_2.setText(loc)
    else:
        self.other_2.setText("")
        self.history_info.setText("")


def prefill(self):
    """Pre-fill sent RST"""
    if len(self.sent.text()) == 0:
        self.sent.setText("599")


def points(self):
    """Return QSO point value"""
    if self.contact_is_dupe > 0:
        return 0
    return 2


def show_mults(self):
    """Return multiplier count: distinct 4-char locator prefixes per band"""
    sql = (
        "select count(DISTINCT(SUBSTR(Exchange1, 1, 4) || ':' || Band)) as mult_count "
        f"from dxlog where ContestNR = {self.database.current_contest} "
        "and typeof(Exchange1) = 'text' "
        "and LENGTH(Exchange1) >= 4;"
    )
    result = self.database.exec_sql(sql)
    if result:
        return int(result.get("mult_count", 0))
    return 0


def show_qso(self):
    """Return QSO count"""
    result = self.database.fetch_qso_count()
    if result:
        return int(result.get("qsos", 0))
    return 0


def calc_score(self):
    """Return calculated score"""
    _points = get_points(self)
    _mults = show_mults(self)
    return _points * _mults


def adif(self):
    """Generate ADIF export"""
    gen_adif(self, cabrillo_name)


def output_cabrillo_line(line_to_output, ending, file_descriptor, file_encoding):
    """Write one encoded line to the Cabrillo file"""
    print(
        line_to_output.encode(file_encoding, errors="ignore").decode(),
        end=ending,
        file=file_descriptor,
    )


def cabrillo(self, file_encoding):
    """Generate Cabrillo log file"""
    logger.debug("******Cabrillo*****")
    logger.debug("Station: %s", f"{self.station}")
    logger.debug("Contest: %s", f"{self.contest_settings}")
    now = datetime.datetime.now()
    date_time = now.strftime("%Y-%m-%d_%H-%M-%S")
    filename = (
        str(Path.home())
        + "/"
        + f"{self.station.get('Call', '').upper().replace('/', '-')}_{cabrillo_name}_{date_time}.log"
    )
    logger.debug("%s", filename)
    log = self.database.fetch_all_contacts_asc()
    try:
        with open(filename, "w", encoding=file_encoding, newline="") as file_descriptor:
            output_cabrillo_line(
                "START-OF-LOG: 3.0", "\r\n", file_descriptor, file_encoding
            )
            output_cabrillo_line(
                f"CREATED-BY: Not1MM v{__version__}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CONTEST: {cabrillo_name}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.station.get("Club", ""):
                output_cabrillo_line(
                    f"CLUB: {self.station.get('Club', '').upper()}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"CALLSIGN: {self.station.get('Call', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-OPERATOR: {self.contest_settings.get('OperatorCategory', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-ASSISTED: {self.contest_settings.get('AssistedCategory', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-BAND: {self.contest_settings.get('BandCategory', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-MODE: {self.contest_settings.get('ModeCategory', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.contest_settings.get("OverlayCategory", "") != "N/A":
                output_cabrillo_line(
                    f"CATEGORY-OVERLAY: {self.contest_settings.get('OverlayCategory', '')}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"GRID-LOCATOR: {self.station.get('GridSquare', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-POWER: {self.contest_settings.get('PowerCategory', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CLAIMED-SCORE: {calc_score(self)}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            ops = ""
            list_of_ops = self.database.get_ops()
            for op in list_of_ops:
                ops += f"{op.get('Operator', '')}, "
            if self.station.get("Call", "") not in ops:
                ops += f"@{self.station.get('Call', '')}"
            else:
                ops = ops.rstrip(", ")
            output_cabrillo_line(
                f"OPERATORS: {ops}", "\r\n", file_descriptor, file_encoding
            )
            output_cabrillo_line(
                f"NAME: {self.station.get('Name', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('Street1', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('City', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('Zip', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('Country', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('Email', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            for contact in log:
                the_date_and_time = contact.get("TS", "")
                themode = contact.get("Mode", "")
                if themode in ("CW-U", "CW-L", "CW-R", "CWR"):
                    themode = "CW"
                if themode in ("LSB", "USB", "AM"):
                    themode = "PH"
                if themode in (
                    "FT8",
                    "FT4",
                    "RTTY",
                    "PSK31",
                    "FSK441",
                    "MSK144",
                    "JT65",
                    "JT9",
                    "Q65",
                ):
                    themode = "DG"
                frequency = str(round(contact.get("Freq", 0))).rjust(5)
                loggeddate = the_date_and_time[:10]
                loggedtime = the_date_and_time[11:13] + the_date_and_time[14:16]
                sent_nr = str(contact.get("SentNr", "0")).zfill(3)
                rcv_nr = str(contact.get("NR", "0")).zfill(3)
                sent_loc = contact.get("GridSquare", "")
                rcv_loc = contact.get("Exchange1", "")
                output_cabrillo_line(
                    f"QSO: {frequency} {themode} {loggeddate} {loggedtime} "
                    f"{contact.get('StationPrefix', '').ljust(13)} "
                    f"{str(contact.get('SNT', '')).ljust(3)} "
                    f"{sent_nr} "
                    f"{sent_loc.ljust(6)} "
                    f"{contact.get('Call', '').ljust(13)} "
                    f"{str(contact.get('RCV', '')).ljust(3)} "
                    f"{rcv_nr} "
                    f"{rcv_loc}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line("END-OF-LOG:", "\r\n", file_descriptor, file_encoding)
        self.show_message_box(f"Cabrillo saved to: {filename}")
    except IOError as exception:
        logger.critical("cabrillo: IO error: %s, writing to %s", exception, filename)
        self.show_message_box(f"Error saving Cabrillo: {exception} {filename}")
        return


def recalculate_mults(self):
    """Walk all contacts and flag the first occurrence of each locator-prefix+Band as a mult"""
    all_contacts = self.database.fetch_all_contacts_asc()
    seen = set()
    for contact in all_contacts:
        exch = contact.get("Exchange1", "")
        band = contact.get("Band", "")
        if len(exch) >= 4:
            key = exch[:4].upper() + ":" + band
            if key not in seen:
                seen.add(key)
                contact["IsMultiplier1"] = 1
            else:
                contact["IsMultiplier1"] = 0
        else:
            contact["IsMultiplier1"] = 0
        self.database.change_contact(contact)


def _validate_field(widget, valid):
    """Color widget red if invalid, reset otherwise"""
    if valid:
        widget.setStyleSheet("")
    else:
        widget.setStyleSheet("background-color: #5c0000;")


def _check_rst(value, rst_min=111, rst_max=599):
    """Return True if value is a valid RST in range"""
    try:
        v = int(value)
        return rst_min <= v <= rst_max
    except ValueError:
        return False


def _check_nr(value):
    """Return True if value is a valid serial number"""
    try:
        v = int(value)
        return 1 <= v <= 999
    except ValueError:
        return False


def _check_locator(value):
    """Return True if value is a valid 6-char Swedish contest locator"""
    if len(value) != 6:
        return False
    prefix = value[:4].upper()
    suffix = value[4:].upper()
    return prefix in VALID_LOCATORS and suffix.isalpha()


def process_esm(self, new_focused_widget=None, with_enter=False):
    """ESM State Machine"""

    # self.pref["run_state"]

    # -----===== Assigned F-Keys =====-----
    # self.esm_dict["CQ"]
    # self.esm_dict["EXCH"]
    # self.esm_dict["QRZ"]
    # self.esm_dict["AGN"]
    # self.esm_dict["HISCALL"]
    # self.esm_dict["MYCALL"]
    # self.esm_dict["QSOB4"]

    # ----==== text fields ====----
    # self.callsign
    # self.sent
    # self.receive
    # self.other_1
    # self.other_2

    if new_focused_widget is not None:
        self.current_widget = self.inputs_dict.get(new_focused_widget)

    if self.current_widget == "callsign":
        self._exch_sent = False

    _validate_field(self.sent, not self.sent.text() or _check_rst(self.sent.text()))
    _validate_field(
        self.receive, not self.receive.text() or _check_rst(self.receive.text())
    )
    _validate_field(
        self.other_1, not self.other_1.text() or _check_nr(self.other_1.text())
    )
    _validate_field(
        self.other_2, not self.other_2.text() or _check_locator(self.other_2.text())
    )

    for a_button in [
        self.F1,
        self.F2,
        self.F3,
        self.F4,
        self.F5,
        self.F6,
        self.F7,
        self.F8,
        self.F9,
        self.F10,
        self.F11,
        self.F12,
    ]:
        self.restore_button_color(a_button)

    buttons_to_send = []

    if self.pref.get("run_state"):
        if self.current_widget == "callsign":
            if len(self.callsign.text()) < 3:
                self.make_button_green(self.esm_dict["CQ"])
                buttons_to_send.append(self.esm_dict["CQ"])
            elif self.contact_is_dupe:
                self.make_button_green(self.esm_dict["QSOB4"])
                buttons_to_send.append(self.esm_dict["QSOB4"])
            elif len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["HISCALL"])
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["HISCALL"])
                buttons_to_send.append(self.esm_dict["EXCH"])

        elif self.current_widget == "other_1":
            if self.other_1.text() == "":
                self.make_button_green(self.F11)
                buttons_to_send.append(self.F11)
            else:
                if with_enter is True:
                    self.other_2.setFocus()
                    return

        elif self.current_widget == "other_2":
            if self.other_2.text() == "":
                self.make_button_green(self.F12)
                buttons_to_send.append(self.F12)
            else:
                self.make_button_green(self.esm_dict["QRZ"])
                buttons_to_send.append(self.esm_dict["QRZ"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        if (
                            not _check_rst(self.sent.text())
                            or not _check_rst(self.receive.text())
                            or not _check_nr(self.other_1.text())
                            or not _check_locator(self.other_2.text())
                        ):
                            return
                        self.save_contact()
                        continue
                    self.process_function_key(button)
            if self.contact_is_dupe:
                return
            if self.current_widget == "callsign" and len(self.callsign.text()) > 2:
                self.other_1.setFocus()
    else:
        if self.current_widget == "callsign":
            if self.contact_is_dupe:
                self.make_button_green(self.esm_dict["QSOB4"])
                buttons_to_send.append(self.esm_dict["QSOB4"])
            elif len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["MYCALL"])
                buttons_to_send.append(self.esm_dict["MYCALL"])

        elif self.current_widget == "other_1":
            if self.other_1.text() == "":
                self.make_button_green(self.F11)
                buttons_to_send.append(self.F11)
            else:
                if with_enter is True:
                    self.other_2.setFocus()
                    return

        elif self.current_widget == "other_2":
            if self.other_2.text() == "":
                self.make_button_green(self.F12)
                buttons_to_send.append(self.F12)
            elif not self._exch_sent:
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["EXCH"])
                if with_enter is True:
                    self._exch_sent = True
            else:
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        if (
                            not _check_rst(self.sent.text())
                            or not _check_rst(self.receive.text())
                            or not _check_nr(self.other_1.text())
                            or not _check_locator(self.other_2.text())
                        ):
                            return
                        self.save_contact()
                        continue
                    self.process_function_key(button)
            if self.current_widget == "callsign":
                self.other_1.setFocus()


def populate_history_info_line(self):
    """Fill history info label from call history"""
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        name = result.get("Name", "") or ""
        loc = result.get("Loc1", "") or ""
        usertext = result.get("UserText", "") or ""
        info = ", ".join(x for x in [result.get("Call", ""), name, loc, usertext] if x)
        self.history_info.setText(info)
    else:
        self.history_info.setText("")


def check_call_history(self):
    """Pre-fill locator from call history if available"""
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        self.history_info.setText(f"{result.get('UserText', '')}")
        if self.other_2.text() == "":
            self.other_2.setText(f"{result.get('Loc1', '')}")


def get_mults(self):
    """Return mult dict for RTC XML"""
    mults = {}
    mults["locator"] = show_mults(self)
    return mults


def just_points(self):
    """Return raw points for RTC XML"""
    return get_points(self)


def imp_adif(self):
    """Import ADIF log"""
    imp_adif(self)



from pathlib import Path
from not1mm.lib.parse_udc import UDC

user_defined_contest = UDC()

path = Path.home() / "UDC_FILES"
path.iterdir()
try:
    print(f"{user_defined_contest.get_udc_names(path)=}")
except Exception as e:
    print(f"{e=}")



# coded and tested by Michael Hartje, DK5HH
# Version 20260318

"""
AGCW qrp, CW
Status:	Active
Member Focus:
Participation:	Worldwide
Awards:	Worldwide
Mode:	CW
Bands:
        80 m (3.510 – 3.560 kHz)
        40 m (7.010 – 7.040 kHz)
        20 m (14.000 – 14.060 kHz)
        15 m (21.000 – 21.070 kHz)
        10 m (28.000 – 28.070 kHz)
Classes:	VLP: Very Low Power = bis 1 W Output
            QRP: klassisch QRP = bis 5 W Output
            MP: moderate Power = bis 25 W Output
            QRO: über 25 W Output

Exchange:   RST+ fortlaufende 3-stellige QSO-Nummer (ab 001) Klasse AGCW-Nummer.
            Beispiel: 579 001 QRP 1537
            Nichtmitglieder geben statt der AGCW-Mitgliedsnummer “NM” für “No Member”.
            Beispiel: 599 027 MP NMember: RST + Serial No + / Mitgliedsnummer
            non-Member: RST + Serial No.

Work stations:	Once per band

QSO Points:	QRO <–> QRO: 0 Points
            QRO <–> QRP, MP, VLP: 2 Points
            MP <–> MP, QRP, VLP: 2 Points
            QRP <–> QRP, VLP: 3 Points
            VLP <–> VLP: 3 Points

Multipliers:	QSO with Member  once per band

Score Calculation:	Total score = total QSO points x total mults

Cabrillo name:	AGCW-QRP
Cabrillo name aliases:	AGCW-QRP

1. Objective	The AGCW-DL invites all radio amateurs to participate in the annual QRP CONTEST. The QRP CONTEST is to arouse and to promote the interest in all aspects of low power amateur radio communication. QRO stations who want to learn about these activities by contacting QRP stations are therefore also invited. They will be counted in a separate category.
2. Dates	Annually on the second Saturday in March from 1400 to 2000 UTC.
3. Bands	80m (3.510-3.560 kHz), 40m, 20m (14.000-14.060 kHz), 15m, 10m. The IARU Region 1 band plan must be adhered to!
4. Categories	VLP: Very Low Power = up to 1 W output
QRP: classic QRP = up to 5 W output
MP: moderate Power = up to 25 W output
QRO: more than 25 W output

In the contest-dialog set PowerCategory to High, Low or QRP but set exchange string for the
exchange text to the  (VLP, QRP, MD, QRO)


5. Operation

Only CW (A1A). Only single OP. Only one TX and RX or TRX is allowed to be operated simultaneously.  Each station may be worked once per band. The use of decoders or reader software is not allowed.

    CALL: CQ QRP TEST
    EXCHANGE:
    RST report plus a sequential 3-digit serial number (starting at 001) followed by category and AGCW-Membership-Number.  Example:599 003 qrp 1234
    Non-members have to send “NM” instead of the membership number. Example: 599 027 MP NM

6. Scoring

    QSO POINTS:
    QRO <–> QRO: 0 points
    QRO <–> QRP, MP, VLP: 2 points
    MP  <–> MP, QRP, VLP: 2 points
    QRP <–> QRP, VLP: 3 points
    VLP <–> VLP: 3 points
    MULTIPLIERS: Each contact with a member of AGCW is worth one multiplier per band.
    TOTAL SCORE: Sum of the QSO points of all bands multiplied with the sum of all multiplier points of all bands.
    Dupes do not count.


7. Awards	Participants can get their own certificate from our website a few weeks after the contest has finished.
At once the results are calculated participants get noticed thereof.

8. Log-Submission



Even short logs are welcome! Final results will be published later.

    LOGS: Please use only Logs in Cabrillo v2 or v3  format. The log should be *uploaded* (URL below) with the following file names: CALL.CBR (e.g. ON3ABC.CBR)
    Participants are requested to make themselves familiar with the Cabrillo specifications beforehand.



    Loglines (QSO-data) within the files must contain sets of fields strictly in this order:

    1. Frequency
    2. Mode (always CW)
    3. Date
    4. Time
    5. Own Call
    6. RST sent
    7. Exchange sent
    8. Other station’s call sign
    9. RST received
    10. Exchange received.

    RST and Exchange , e.g. 599 001 QRP 1234 or
    599 003 VLP NM

    A logging program was provided by ArComm and can be downloaded as freeware here:
    http://www.qslonline.de/hk/eigen/kontest.htm#hamagcwqrp
    Log Submission
    New from 2023: Submission exclusively via web upload on the page
    https://contest.agcw.de/qrpc/


    DEADLINE: Upload no later than March, 31.

Checklogs are welcome, also comments of the participants.
Publication of the claimed scores will be announced promptly after the deadline.
"""

import datetime
import logging
import platform
import re

from pathlib import Path

from PyQt6 import QtWidgets

from not1mm.lib.plugin_common import gen_adif, imp_adif, get_points, online_score_xml

from not1mm.lib.version import __version__

logger = logging.getLogger(__name__)

EXCHANGE_HINT = "# pwr_class MemberNr"
SOAPBOX_HINT = """
Power classes are {VLP, QRP, MP, QRO}
MemberNr is AGCW-MemberNr or NM for Non Member.

For the Run exchange macro I’d put ’{SNT} # {EXCH}’"""


name = "AGCW QRP"
cabrillo_name = "AGCW-QRP"
mode = "CW"  # CW SSB BOTH RTTY

columns = [
    "YYYY-MM-DD HH:MM:SS",
    "Call",
    "Freq",
    "Mode",
    "Snt",
    "Rcv",
    "SentNr",
    "RcvNr",
    "M1",
    "PTS",
]

advance_on_space = [True, True, True, True, False]

# 1 once per contest, 2 work each band, 3 each band/mode, 4 no dupe checking
dupe_type = 2


def init_contest(self):
    """setup plugin"""
    set_tab_next(self)
    set_tab_prev(self)
    interface(self)
    self.next_field = self.other_2


def interface(self):
    """Setup user interface"""
    self.field1.show()
    self.field2.show()
    self.field3.show()
    self.field4.show()
    self.other_label.setText("Sent")
    self.other_1.setAccessibleName("Sent")
    self.exch_label.setText("SN Pwr MbrNr")
    self.other_2.setAccessibleName("Serial Number Power Class Member Number")


def reset_label(self):
    """reset label after field cleared"""


def set_tab_next(self):
    """Set TAB Advances"""
    self.tab_next = {
        self.callsign: self.other_1,
        self.sent: self.other_1,
        self.receive: self.other_1,
        self.other_1: self.other_2,
        #        self.other_2: self.other_3,
        self.other_2: self.callsign,
    }


def set_tab_prev(self):
    """Set TAB Advances"""
    self.tab_prev = {
        self.callsign: self.other_2,
        self.sent: self.callsign,
        self.receive: self.callsign,
        self.other_1: self.callsign,
        self.other_2: self.other_1,
        #        self.other_3: self_other_2
    }


def set_contact_vars(self):
    """
    prepare both multipliers even when only 0ne is used
    """
    self.contact["SNT"] = self.sent.text()
    self.contact["RCV"] = self.receive.text()
    self.contact["SentNr"] = self.other_1.text().upper() + self.contest_settings.get(
        "SentExchange", "xxx xxx"
    ).strip("#")
    self.contact["NR"] = self.other_2.text().upper().replace(" ", "/")

    self.contact["IsMultiplier1"] = 0
    self.contact["IsMultiplier2"] = 0

    band = self.contact.get("Band", "")

    if len(self.contact["NR"]) > 0:
        r_number, pwr_class, member = received_fields(self, self.contact["NR"])
        if member.isdigit():
            query = (
                f"select count(*) as member_count from dxlog where Nr like '%/%/{member}' "
                f"and Band = '{band}' "
                f"and ContestNR = {self.pref.get('contest', '1')};"
            )
            result = self.database.exec_sql(query)
            count = int(result.get("member_count", 0))
            if count == 0:
                self.contact["IsMultiplier1"] = 1


def predupe(self):
    """called after callsign entered"""


def prefill(self):
    """Fill SentNR"""
    sent_sxchange_setting = self.contest_settings.get("SentExchange", "")
    if (len(sent_sxchange_setting) > 0) and (sent_sxchange_setting.find("#") >= 0):
        if self.current_sn is not None:
            serial_nr = str(self.current_sn).zfill(3)
        if serial_nr == "None":
            serial_nr = "001"
        if len(self.other_1.text()) == 0:
            self.other_1.setText(serial_nr)
    else:
        self.other_1.setText(sent_sxchange_setting)


def received_fields(self, r_nr) -> [str, str, str]:
    """
    split the "received field" 003/VLP/1234
    """
    hs = r_nr.upper().replace("/", "")
    if len(hs) > 3:  # there could be a valid message inside
        pwr_class = re.findall("[A-Z]+[OP]", hs)[0]
        hr = hs.split(pwr_class)
        member = hr[1]
        r_number = hr[0]
    else:
        member = "NM"
        pwr_class = "QRO"
        r_number = "001"
    return r_number, pwr_class, member


def points(self):
    """
    from contest rules:
    QSO Points:
            QRO <–> QRO: 0 Points

            QRO <–> QRP, MP, VLP: 2 Points

            MP <–> MP, QRP, VLP: 2 Points

            QRP <–> QRP, VLP: 3 Points

            VLP <–> VLP: 3 Punkte1 point per QSO

    """
    # The Dialog offers QRP, Low, High for the powercategory
    # so we code it
    #    VLP and QRP will be coded as QRP,
    #    MP will be coded Low
    #    QRO will be coded as High

    mypowerclass = self.contest_settings.get("PowerCategory")

    if self.contact_is_dupe > 0:
        return 0
    pts = 2  # for qrp stns
    if len(self.contact["NR"]) > 0:
        r_number, pwr_class, member = received_fields(self, self.contact["NR"])
        pt_2 = {"QRO", "MP"}
        pt_3 = {"QRP", "VLP"}

        if pwr_class in pt_2:
            pts = 2
        if pwr_class in pt_3:
            pts = 3

        self.contact["NR"] = r_number + "/" + pwr_class + "/" + member
    # if both station QRO than pts=0
    if mypowerclass == "High" and pwr_class == "QRO":
        pts = 0
    return pts


def show_mults(self, rtc=None):
    """Return display string for mults"""
    one = int(self.database.fetch_mult_count(1).get("count", 0))
    two = int(self.database.fetch_mult_count(2).get("count", 0))
    if rtc is not None:
        return (two, one)

    return one + two


def show_qso(self):
    """Return qso count"""
    result = self.database.fetch_qso_count()
    if result:
        return int(result.get("qsos", 0))
    return 0


def calc_score(self):
    """Return calculated score"""
    result = self.database.fetch_points()
    if result is not None:
        score = result.get("Points", "0")
        if score is None:
            score = "0"
        contest_points = int(score)
        mults = show_mults(self)
        return contest_points * mults
    return 0


def recalculate_mults(self):
    """Recalculates multipliers after change in logged qso."""

    all_contacts = self.database.fetch_all_contacts_asc()
    for contact in all_contacts:

        contact["IsMultiplier1"] = 0
        contact["IsMultiplier2"] = 0

        time_stamp = contact.get("TS", "")
        R_NR = contact.get("NR", "")
        # dxcc = contact.get("CountryPrefix", "")
        band = contact.get("Band", "")
        _, _, member = received_fields(self, R_NR)
        if (member != "NM") and member.isdigit():
            query = (
                f"select count(*) as member_count from dxlog where  TS < '{time_stamp}' "
                f"and NR like '%/%/{member}' "
                f"and Band = '{band}' "
                f"and ContestNR = {self.pref.get('contest', '1')};"
            )
            result = self.database.exec_sql(query)
            count = int(result.get("member_count", 0))
            if count == 0:
                contact["IsMultiplier1"] = 1

        # use the following code for other agcw-contests e.g happy new year contest
        """if dxcc:
            query = (
                f"select count(*) as dxcc_count from dxlog where TS < '{time_stamp}' "
                f"and CountryPrefix = '{dxcc}' "
                f"and Band = '{band}' "
                f"and ContestNR = {self.pref.get('contest', '1')};"
            )
            result = self.database.exec_sql(query)
            if not result.get("dxcc_count", ""):
                contact["IsMultiplier2"] = 1
        """
        self.database.change_contact(contact)
    cmd = {}
    cmd["cmd"] = "UPDATELOG"
    if self.log_window:
        self.log_window.msg_from_main(cmd)


def adif(self):
    """Call the generate ADIF function"""
    gen_adif(self, cabrillo_name, "AGCW-QRP")


def output_cabrillo_line(line_to_output, ending, file_descriptor, file_encoding):
    """"""
    print(
        line_to_output.encode(file_encoding, errors="ignore").decode(),
        end=ending,
        file=file_descriptor,
    )


def cabrillo(self, file_encoding):
    """Generates Cabrillo file. Maybe."""
    # https://www.cqwpx.com/cabrillo.htm
    logger.debug("******Cabrillo*****")
    logger.debug("Station: %s", f"{self.station}")
    logger.debug("Contest: %s", f"{self.contest_settings}")
    now = datetime.datetime.now()
    date_time = now.strftime("%Y-%m-%d_%H-%M-%S")
    filename = (
        str(Path.home())
        + "/"
        + f"{self.station.get('Call', '').upper().replace('/','-')}_{cabrillo_name}_{date_time}.log"
    )
    logger.debug("%s", filename)
    log = self.database.fetch_all_contacts_asc()

    try:
        with open(filename, "w", encoding=file_encoding, newline="") as file_descriptor:
            output_cabrillo_line(
                "START-OF-LOG: 3.0",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CREATED-BY: Not1MM v{__version__}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CONTEST: {cabrillo_name}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.station.get("Club", ""):
                output_cabrillo_line(
                    f"CLUB: {self.station.get('Club', '').upper()}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"CALLSIGN: {self.station.get('Call','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"LOCATION: {self.station.get('ARRLSection', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-OPERATOR: {self.contest_settings.get('OperatorCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-ASSISTED: {self.contest_settings.get('AssistedCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-BAND: {self.contest_settings.get('BandCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            mode = self.contest_settings.get("ModeCategory", "")
            if mode in ["SSB+CW", "SSB+CW+DIGITAL"]:
                mode = "MIXED"
            output_cabrillo_line(
                f"CATEGORY-MODE: {mode}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-TRANSMITTER: {self.contest_settings.get('TransmitterCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.contest_settings.get("OverlayCategory", "") != "N/A":
                output_cabrillo_line(
                    f"CATEGORY-OVERLAY: {self.contest_settings.get('OverlayCategory','')}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"GRID-LOCATOR: {self.station.get('GridSquare','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-POWER: {self.contest_settings.get('PowerCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )

            output_cabrillo_line(
                f"CLAIMED-SCORE: {calc_score(self)}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            ops = ""
            list_of_ops = self.database.get_ops()
            for op in list_of_ops:
                ops += f"{op.get('Operator', '')}, "
            if self.station.get("Call", "") not in ops:
                ops += f"@{self.station.get('Call','')}"
            else:
                ops = ops.rstrip(", ")
            output_cabrillo_line(
                f"OPERATORS: {ops}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"NAME: {self.station.get('Name', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('Street1', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-CITY: {self.station.get('City', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-STATE-PROVINCE: {self.station.get('State', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-POSTALCODE: {self.station.get('Zip', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-COUNTRY: {self.station.get('Country', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"EMAIL: {self.station.get('Email', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.contest_settings.get("Soapbox") != SOAPBOX_HINT:
                for l in self.contest_settings.get("Soapbox", "").splitlines():
                    output_cabrillo_line(
                        f"SOAPBOX: {l}",
                        "\r\n",
                        file_descriptor,
                        file_encoding,
                    )

            for contact in log:
                the_date_and_time = contact.get("TS", "")
                themode = contact.get("Mode", "")
                if themode == "LSB" or themode == "USB":
                    themode = "PH"
                if themode in ("CW-U", "CW-L", "CW-R", "CWR"):
                    themode = "CW"
                frequency = str(round(contact.get("Freq", "0"))).rjust(5)

                loggeddate = the_date_and_time[:10]
                loggedtime = the_date_and_time[11:13] + the_date_and_time[14:16]
                output_cabrillo_line(
                    f"QSO: {frequency} {themode} {loggeddate} {loggedtime} "
                    f"{contact.get('StationPrefix', '').ljust(13)} "
                    f"{str(contact.get('SNT', '')).ljust(3)} "
                    f"{str(contact.get('SentNr', '')).ljust(6)} "
                    f"{contact.get('Call', '').ljust(13)} "
                    f"{str(contact.get('RCV', '')).ljust(3)} "
                    f"{str(contact.get('NR', '')).ljust(6).replace('/',' ')}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line("END-OF-LOG:", "\r\n", file_descriptor, file_encoding)
        self.show_message_box(f"Cabrillo saved to: {filename}")
    except IOError as exception:
        logger.critical("cabrillo: IO error: %s, writing to %s", exception, filename)
        self.show_message_box(f"Error saving Cabrillo: {exception} {filename}")
        return


def process_esm(self, new_focused_widget=None, with_enter=False):
    """ESM State Machine"""

    # self.pref["run_state"]

    # -----===== Assigned F-Keys =====-----
    # self.esm_dict["CQ"]
    # self.esm_dict["EXCH"]
    # self.esm_dict["QRZ"]
    # self.esm_dict["AGN"]
    # self.esm_dict["HISCALL"]
    # self.esm_dict["MYCALL"]
    # self.esm_dict["QSOB4"]

    # ----==== text fields ====----
    # self.callsign
    # self.sent
    # self.receive
    # self.other_1
    # self.other_2

    if new_focused_widget is not None:
        self.current_widget = self.inputs_dict.get(new_focused_widget)

    # print(f"checking esm {self.current_widget=} {with_enter=} {self.pref.get("run_state")=}")

    for a_button in [
        self.esm_dict["CQ"],
        self.esm_dict["EXCH"],
        self.esm_dict["QRZ"],
        self.esm_dict["AGN"],
        self.esm_dict["HISCALL"],
        self.esm_dict["MYCALL"],
        self.esm_dict["QSOB4"],
    ]:
        if a_button is not None:
            self.restore_button_color(a_button)

    buttons_to_send = []

    if self.pref.get("run_state"):
        if self.current_widget == "callsign":
            if len(self.callsign.text()) < 3:
                self.make_button_green(self.esm_dict["CQ"])
                buttons_to_send.append(self.esm_dict["CQ"])
            elif len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["HISCALL"])
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["HISCALL"])
                buttons_to_send.append(self.esm_dict["EXCH"])

        elif self.current_widget == "other_2":
            if self.other_2.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["QRZ"])
                buttons_to_send.append(self.esm_dict["QRZ"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    self.process_function_key(button)
    else:
        if self.current_widget == "callsign":
            if len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["MYCALL"])
                buttons_to_send.append(self.esm_dict["MYCALL"])

        elif self.current_widget == "other_2":
            if self.other_2.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["EXCH"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    self.process_function_key(button)


def get_mults(self):
    """Get mults for RTC XML"""
    mults = {}
    mults["country"], mults["state"] = show_mults(self, rtc=True)
    return mults


def just_points(self):
    """Get points for RTC XML"""
    return get_points(self)


def populate_history_info_line(self):
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        self.history_info.setText(
            f"{result.get('Call', '')}, {result.get('Exch1', '')}, {result.get('UserText','...')}"
        )
    else:
        self.history_info.setText("")


def check_call_history(self):
    """"""
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        self.history_info.setText(f"{result.get('UserText','')}")
        if self.other_2.text() == "":
            self.other_2.setText(f"{result.get('Exch1', '')}")



import logging

if __name__ == "__main__":
    print("I'm not the program you are looking for.")

logger = logging.getLogger("parse_udc")


class UDC:
    """Parse UDC files"""

    # Sections
    # [Author]
    # [File]
    # [Contest]

    def __init__(self) -> None: ...

    def parse_udc(self, path: str) -> dict[str, dict[str, str]]:
        """Parse a .udc file at `path` and return a dict of sections.

        Each section is a dict of key -> value (both stripped). Lines without
        an '=' are ignored. Empty values are preserved as empty strings.
        """
        result: dict[str, dict[str, str]] = {}
        current_section: str | None = None

        with open(path, "r", encoding="utf-8") as fh:
            for raw in fh:
                line = raw.strip()
                if not line:
                    continue
                if line.startswith("[") and line.endswith("]"):
                    current_section = line[1:-1].strip()
                    if current_section not in result:
                        result[current_section] = {}
                    continue
                if current_section is None:
                    # skip any lines before first section
                    continue

                # split on first '='; keys or values may contain '=' rarely
                if "=" not in line:
                    continue
                key, _, value = line.partition("=")
                key = key.strip()
                value = value.strip()
                result[current_section][key] = value

        return result



from PyQt6.QtWidgets import (
    QDockWidget,
    QGraphicsScene,
    QGraphicsPathItem,
    QGraphicsPixmapItem,
)

from PyQt6.QtGui import (
    QImage,
    QColor,
    QPixmap,
    QPen,
    QBrush,
    QPainterPath,
    QShowEvent,
    QResizeEvent,
    QMouseEvent,
)

from PyQt6.QtCore import Qt, pyqtSignal, QTimer
from PyQt6 import uic

from not1mm.lib.rot_interface import RotatorInterface
import not1mm.fsutils as fsutils
import math
import logging
import os

logger = logging.getLogger(__name__)


class RotatorWindow(QDockWidget):
    message: pyqtSignal = pyqtSignal(dict)
    pref: dict = {}
    MAP_RESOLUTION: int = 600
    GLOBE_RADIUS: float = 100.0
    requestedAzimuthNeedle: QGraphicsPathItem | None = None
    antennaNeedle: QGraphicsPathItem | None = None
    rotatorwindow_closed = pyqtSignal()

    def __init__(self, action, host: str = "127.0.0.1", port: int = 4533):
        super().__init__()
        self.action = action
        self.host = host
        self.port = port
        self.active: bool = False
        self.compassScene: QGraphicsScene | None = None
        self.mygrid: str = "DM13at"
        self.requestedAzimuth: float | None = None
        self.antennaAzimuth: float | None = None
        uic.loadUi(fsutils.APP_DATA_PATH / "rotator.ui", self)
        self.north_button.clicked.connect(self.set_north_azimuth)
        self.south_button.clicked.connect(self.set_south_azimuth)
        self.east_button.clicked.connect(self.set_east_azimuth)
        self.west_button.clicked.connect(self.set_west_azimuth)
        self.move_button.clicked.connect(self.the_eye_of_sauron) # left-click
        self.move_button.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.move_button.customContextMenuRequested.connect(self.rotate_long_path) # right-click
        self.stop_button.clicked.connect(self.stop)
        self.park_button.clicked.connect(lambda x: self.rotator.send_command("K"))
        self.redrawMap()
        self.rotator: RotatorInterface = RotatorInterface(self.host, self.port)
        self.antennaAzimuth, _ = self.rotator.get_position()
        self.set_antenna_azimuth(self.antennaAzimuth)
        self.watch_timer: QTimer = QTimer()
        self.watch_timer.timeout.connect(self.check_rotator)
        self.watch_timer.start(1000)

    def set_host_port(self, host: str, port: int) -> None:
        """"""
        self.host = host
        self.port = port
        self.rotator.set_host_port(self.host, self.port)
        self.redrawMap()

    def msg_from_main(self, msg: dict) -> None:
        """"""
        if self.active is True and isinstance(msg, dict):
            if msg.get("cmd", "") in ("UPDATELOG", "CONTACTCHANGED", "DELETED"):
                ...
            if msg.get("cmd", "") == "NEWDB":
                ...

    def set_mygrid(self, mygrid: str) -> None:
        """"""
        if isinstance(mygrid, str):
            self.mygrid = mygrid
            self.redrawMap()

    def setActive(self, active: bool) -> None:
        """"""
        if isinstance(active, bool):
            self.active = active

    def set_requested_azimuth(self, azimuth: float) -> None:
        if isinstance(azimuth, float):
            self.requestedAzimuth = azimuth
            self.requestedAzimuthNeedle.setRotation(self.requestedAzimuth)
            self.requestedAzimuthNeedle.show()
        else:
            self.requestedAzimuthNeedle.hide()

    def set_antenna_azimuth(self, azimuth: float) -> None:
        if isinstance(azimuth, float):
            self.setWindowTitle(f"Rotator: {azimuth:.0f}°")
            self.antennaAzimuth = azimuth
            self.antennaNeedle.setRotation(self.antennaAzimuth)
            self.antennaNeedle.show()
        else:
            self.setWindowTitle("Rotator")
            self.antennaNeedle.hide()

    def set_north_azimuth(self) -> None:
        """Point the antenna North."""
        if self.rotator.connected:
            self.rotator.set_position(0.0)

    def set_south_azimuth(self) -> None:
        """Point the antenna South."""
        if self.rotator.connected:
            self.rotator.set_position(180.0)

    def set_east_azimuth(self) -> None:
        """Point the antenna East."""
        if self.rotator.connected:
            self.rotator.set_position(90.0)

    def set_west_azimuth(self) -> None:
        """Point the antenna West."""
        if self.rotator.connected:
            self.rotator.set_position(270.0)

    def the_eye_of_sauron(self) -> None:
        """Move the antennas azimuth to match the contacts."""
        if self.rotator.connected and self.requestedAzimuth is not None:
            self.rotator.set_position(self.requestedAzimuth)

    def rotate_long_path(self) -> None:
        """Move the antennas azimuth to long-path direction."""
        if self.rotator.connected and self.requestedAzimuth is not None:
            self.rotator.set_position((self.requestedAzimuth + 180.0) % 360.0)

    def stop(self) -> None:
        """Stop the rotator."""
        self.rotator.send_command("S")

    def redrawMap(self) -> None:
        """"""
        self.compassScene: QGraphicsScene = QGraphicsScene()
        self.compassView.setScene(self.compassScene)
        self.compassView.setStyleSheet("background-color: transparent;")
        file = fsutils.APP_DATA_PATH / "map3.png"
        source: QImage = QImage()
        source.load(str(file))
        the_map: QImage = QImage(
            self.MAP_RESOLUTION, self.MAP_RESOLUTION, QImage.Format.Format_ARGB32
        )
        the_map.fill(QColor(0, 0, 0, 0))
        lat, lon = self.gridtolatlon(self.mygrid)

        if os.path.exists(f"{fsutils.USER_DATA_PATH}/{self.mygrid}v2.png"):
            the_map.load(f"{fsutils.USER_DATA_PATH}/{self.mygrid}v2.png")
        else:
            the_map = self.equirectangular_to_azimuthal_equidistant(
                source, lat, lon, output_size=self.MAP_RESOLUTION
            )

        pixMapItem: QGraphicsPixmapItem | None = self.compassScene.addPixmap(
            QPixmap.fromImage(the_map)
        )
        if pixMapItem is None:
            logging.error("Unable to add pixmap to scene")
        else:
            pixMapItem.moveBy(-self.MAP_RESOLUTION / 2, -self.MAP_RESOLUTION / 2)
            pixMapItem.setTransformOriginPoint(
                self.MAP_RESOLUTION / 2, self.MAP_RESOLUTION / 2
            )
            pixMapItem.setScale(self.GLOBE_RADIUS * 2 / self.MAP_RESOLUTION)
        self.compassScene.addEllipse(
            -100,
            -100,
            self.GLOBE_RADIUS * 2,
            self.GLOBE_RADIUS * 2,
            QPen(QColor(100, 100, 100)),
            QBrush(QColor(0, 0, 0), Qt.BrushStyle.NoBrush),
        )
        self.compassScene.addEllipse(
            -1,
            -1,
            2,
            2,
            QPen(Qt.PenStyle.NoPen),
            QBrush(QColor(0, 0, 0), Qt.BrushStyle.SolidPattern),
        )
        path: QPainterPath = QPainterPath()
        path.lineTo(-4, 0)
        path.lineTo(0, -90)
        path.lineTo(4, 0)
        path.closeSubpath()

        # path2: QPainterPath = QPainterPath()
        # path2.lineTo(-1, 0)
        # path2.lineTo(0, -90)
        # path2.lineTo(1, 0)
        # path2.closeSubpath()

        self.requestedAzimuthNeedle: QGraphicsPathItem | None = (
            self.compassScene.addPath(
                path,
                QPen(QColor(0, 0, 0, 150)),
                QBrush(QColor(0, 0, 255), Qt.BrushStyle.SolidPattern),
            )
        )

        if (
            isinstance(self.requestedAzimuth, float)
            and self.requestedAzimuthNeedle is not None
        ):
            self.requestedAzimuthNeedle.setRotation(self.requestedAzimuth)
            self.requestedAzimuthNeedle.show()
        else:
            self.requestedAzimuthNeedle.hide()

        self.antennaNeedle: QGraphicsPathItem | None = self.compassScene.addPath(
            path,
            QPen(QColor(0, 0, 0, 150)),
            QBrush(QColor(255, 191, 0), Qt.BrushStyle.SolidPattern),
        )
        if isinstance(self.antennaAzimuth, float) and self.antennaNeedle is not None:
            self.antennaNeedle.setRotation(self.antennaAzimuth)
            self.antennaNeedle.show()
        else:
            self.antennaNeedle.hide()

    def gridtolatlon(self, maiden: str) -> tuple[float, float]:
        """
        Converts a maidenhead gridsquare to a latitude longitude pair.
        """
        try:
            maiden: str = str(maiden).strip().upper()

            chars_in_grid_square: int = len(maiden)
            if not 8 >= chars_in_grid_square >= 2 and chars_in_grid_square % 2 == 0:
                return 0.0, 0.0
            lon = (ord(maiden[0]) - 65) * 20 - 180
            lat = (ord(maiden[1]) - 65) * 10 - 90
            if chars_in_grid_square >= 4:
                lon += (ord(maiden[2]) - 48) * 2
                lat += (ord(maiden[3]) - 48) * 1
                if chars_in_grid_square >= 6:
                    lon += (ord(maiden[4]) - 65) * (5.0 / 60.0)
                    lat += (ord(maiden[5]) - 65) * (2.5 / 60.0)
                    if chars_in_grid_square >= 8:
                        lon += (ord(maiden[6]) - 48) * (30.0 / 3600.0)
                        lat += (ord(maiden[7]) - 48) * (15.0 / 3600.0)
                        lon += 15.0 / 3600.0
                        lat += 7.5 / 3600.0
                    else:
                        lon += 2.5 / 60.0
                        lat += 1.25 / 60.0
                else:
                    lon += 1
                    lat += 0.5
            else:
                lon += 10
                lat += 5

            return float(lat), float(lon)
        except IndexError:
            return 0.0, 0.0

    def equirectangular_to_azimuthal_equidistant(
        self,
        source_img: QImage,
        center_lat_deg: float,
        center_lon_deg: float,
        output_size: int = 600,
    ) -> QImage:
        """This does some super magic"""

        width, height = source_img.width(), source_img.height()
        dest_img: QImage = QImage(output_size, output_size, QImage.Format.Format_ARGB32)

        # Convert center point to radians
        lat0: float = math.radians(center_lat_deg)
        lon0: float = math.radians(center_lon_deg)

        sin_lat: float = math.sin(lat0)
        cos_lat: float = math.cos(lat0)

        R: float = output_size / 2
        for y in range(output_size):
            for x in range(output_size):
                # Convert pixel coordinate to normalized cartesian coordinate in range [-1, 1]
                dx = (x - R) / R
                dy = (R - y) / R  # Invert Y axis to have +Y upwards

                rho = math.sqrt(dx * dx + dy * dy)
                if rho > 1.0:
                    # Outside the projection circle
                    dest_img.setPixelColor(x, y, QColor(0, 0, 0, 0))
                    continue

                c = (
                    rho * math.pi
                )  # scale rho to angular distance (c = great-circle distance)

                if rho == 0:
                    lat = lat0
                    lon = lon0
                else:
                    sin_c = math.sin(c)
                    cos_c = math.cos(c)

                    lat = math.asin(cos_c * sin_lat + (dy * sin_c * cos_lat / rho))

                    lon = lon0 + math.atan2(
                        dx * sin_c,
                        rho * cos_lat * cos_c - dy * sin_lat * sin_c,
                    )

                # Map lat/lon to source image pixel coordinates (equirectangular)
                lon_deg = math.degrees(lon)
                lat_deg = math.degrees(lat)

                # Wrap longitude to [-180, 180]
                if lon_deg < -180:
                    lon_deg += 360
                elif lon_deg > 180:
                    lon_deg -= 360

                # Map to source image coordinates
                src_x = int(((lon_deg + 180.0) / 360.0) * width)
                src_y = int(((90 - lat_deg) / 180.0) * height)

                # Clamp coordinates
                src_x = max(0, min(width - 1, src_x))
                src_y = max(0, min(height - 1, src_y))

                color = source_img.pixelColor(src_x, src_y)
                dest_img.setPixelColor(x, y, color)

        dest_img.save(f"{fsutils.USER_DATA_PATH}/{self.mygrid}v2.png", "PNG")
        return dest_img

    def showEvent(self, event: QShowEvent) -> None:
        """Make the globe fit in the widget when widget is shown."""
        self.compassView.fitInView(
            self.compassScene.sceneRect(), Qt.AspectRatioMode.KeepAspectRatio
        )

    def resizeEvent(self, event: QResizeEvent) -> None:
        """Make globe fit in widget when widget is resized."""
        self.compassView.fitInView(
            self.compassScene.sceneRect(), Qt.AspectRatioMode.KeepAspectRatio
        )

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Catch mouse clicks in the widget"""
        if event.button() == Qt.MouseButton.LeftButton:
            clickPos = self.compassView.mapToScene(
                self.compassView.mapFromGlobal(event.globalPosition().toPoint())
            )
            dx: float = clickPos.x()
            dy: float = -1 * clickPos.y()
            if math.sqrt(math.pow(dx, 2) + math.pow(dy, 2)) <= self.GLOBE_RADIUS:

                angle: float = math.degrees(math.atan2(dx, dy))

                if angle < 0:
                    angle += 360

                self.rotator.set_position(angle)

    def check_rotator(self) -> None:
        """Check the rotator"""
        if self.active is False:
            return
        if self.rotator.connected:
            self.antennaAzimuth, _ = self.rotator.get_position()
            self.set_antenna_azimuth(self.antennaAzimuth)
        else:
            self.rotator.connect()

    def closeEvent(self, event) -> None:
        self.action.setChecked(False)
        self.rotatorwindow_closed.emit()
        event.accept()



"""
Generic State QSO Party (RST + County)
-------------------------------------------------
Status:	Active
Geographic Focus:	US
Participation:	Worldwide
Mode:	CW, SSB, Digital
Bands:	160, 80, 40, 20, 15, 10, 6, 2m, 70cm
Classes:	Single Op (QRP/Low/High)
            Multi-Single (QRP/Low/High)
            Multi-Multi (QRP/Low/High)
Max operating hours:	Varies
Max power:	HP: >100 watts
LP: 100 watts
QRP: 5 watts
Exchange:	RST + County/State/Country
Work stations:	Once per band per mode per location

Scoring:
Contact via phone: 1 point
Contact via CW: 2 points
Contact via digital: 2 points

Multipliers: County, State, or Country (aside from US/Canada) across all modes/bands: 1 point

Score Calculation:	Total score = total QSO points x total mults + bonus points
E-mail logs to:	(varies)
Mail logs to:	(varies)
Find rules at:	(varies)
Cabrillo name:	QSO_PARTY
"""

import datetime
import logging
import platform

from pathlib import Path

from PyQt6 import QtWidgets

from not1mm.lib.plugin_common import gen_adif, imp_adif, get_points

from not1mm.lib.version import __version__

logger = logging.getLogger(__name__)

EXCHANGE_HINT = "County"

name = "QSO PARTY RST"
cabrillo_name = "QSO_PARTY"
mode = "BOTH"  # CW SSB BOTH RTTY

columns = [
    "YYYY-MM-DD HH:MM:SS",
    "Call",
    "Freq",
    "Mode",
    "Snt",
    "Rcv",
    "Exchange1",
    "M1",
    "PTS",
    "Operator",
]

advance_on_space = [True, True, True, True, True]

# 1 once per contest, 2 work each band, 3 each band/mode, 4 no dupe checking
dupe_type = 4
# allow for rovers


def init_contest(self):
    """setup plugin"""
    set_tab_next(self)
    set_tab_prev(self)
    interface(self)
    self.next_field = self.other_1


def interface(self):
    """Setup user interface"""
    self.field1.show()
    self.field2.show()
    self.field3.show()
    self.field4.hide()
    self.snt_label.setText("Sent RST")
    self.sent.setAccessibleName("Snt")
    self.rcv_label.setText("Rcv RST")
    self.receive.setAccessibleName("Rcv")
    self.other_label.setText("County/State/DX")
    # self.sent.setText("")


def reset_label(self):
    """reset label after field cleared"""


def set_tab_next(self):
    """Set TAB Advances"""
    self.tab_next = {
        self.callsign: self.sent,
        self.sent: self.receive,
        self.receive: self.other_1,
        self.other_1: self.callsign,
    }


def set_tab_prev(self):
    """Set Shift-TAB Advances"""
    self.tab_prev = {
        self.callsign: self.other_1,
        self.sent: self.callsign,
        self.receive: self.sent,
        self.other_1: self.receive,
    }


def set_contact_vars(self):
    """Contest Specific"""
    self.contact["SNT"] = self.sent.text()
    self.contact["RCV"] = self.receive.text()
    self.contact["Exchange1"] = self.other_1.text().upper()

    self.contact["IsMultiplier1"] = 0
    self.contact["IsMultiplier2"] = 0

    county = self.contact.get("Exchange1", "").upper()
    query = (
        f"select count(*) as county_count from dxlog where "
        f"Exchange1 = '{county}' "
        f"and ContestNR = {self.pref.get('contest', '1')};"
    )
    result = self.database.exec_sql(query)
    count = int(result.get("county_count", 0))
    if count == 0:
        self.contact["IsMultiplier1"] = 1


def predupe(self):
    """called after callsign entered"""


def prefill(self):
    """Fill Sent"""
    qso_mode = self.contact.get("Mode", "")
    match qso_mode:
        case "LSB" | "USB" | "SSB" | "FM" | "AM":
            self.sent.setText("599")
            self.rcv.setText("599")
        case "CW" | "CW-U" | "CW-L" | "CW-R" | "CWR":
            self.sent.setText("59")
            self.rcv.setText("59")
        case (
            "FT8"
            | "FT4"
            | "RTTY"
            | "PSK31"
            | "FSK441"
            | "MSK144"
            | "JT65"
            | "JT9"
            | "Q65"
            | "PKTUSB"
            | "PKTLSB"
        ):
            self.sent.setText("599")
            self.rcv.setText("599")
    # end match


def points(self):
    """
    Scoring:
    Phone = 1 point
    CW = 2 points
    Digital = 2 points
    """

    if self.contact_is_dupe > 0:
        return 0

    qso_mode = self.contact.get("Mode", "")
    match qso_mode:
        case "LSB" | "USB" | "SSB" | "FM" | "AM":
            return 1
        case "CW" | "CW-U" | "CW-L" | "CW-R" | "CWR":
            return 2
        case (
            "FT8"
            | "FT4"
            | "RTTY"
            | "PSK31"
            | "FSK441"
            | "MSK144"
            | "JT65"
            | "JT9"
            | "Q65"
            | "PKTUSB"
            | "PKTLSB"
        ):
            return 2
        case _:
            return 0
    # end match


def show_mults(self, rtc=None):
    """Return display string for mults"""
    result = int(self.database.fetch_mult_count(1).get("count", 0))
    return result


def show_qso(self):
    """Return qso count"""
    result = self.database.fetch_qso_count()
    if result:
        return int(result.get("qsos", 0))
    return 0


def calc_score(self):
    """Return calculated score"""
    result = self.database.fetch_points()
    if result is not None:
        score = result.get("Points", "0")
        if score is None:
            score = "0"
        contest_points = int(score)
        mults = show_mults(self)
        return contest_points * mults
    return 0


def recalculate_mults(self):
    """Recalculates multipliers after change in logged qso."""

    all_contacts = self.database.fetch_all_contacts_asc()
    for contact in all_contacts:

        contact["IsMultiplier1"] = 0
        contact["IsMultiplier2"] = 0

        time_stamp = contact.get("TS", "")
        county = contact.get("Exchange1", "")
        query = (
            f"select count(*) as county_count from dxlog where  TS < '{time_stamp}' "
            f"and Exchange1 = '{county.upper()}' "
            f"and ContestNR = {self.pref.get('contest', '1')};"
        )
        result = self.database.exec_sql(query)
        count = int(result.get("county_count", 0))
        if count == 0:
            contact["IsMultiplier1"] = 1

        self.database.change_contact(contact)

    cmd = {}
    cmd["cmd"] = "UPDATELOG"
    if self.log_window:
        self.log_window.msg_from_main(cmd)


def adif(self):
    """Call the generate ADIF function"""
    gen_adif(self, cabrillo_name, "QSO_PARTY")


def output_cabrillo_line(line_to_output, ending, file_descriptor, file_encoding):
    """"""
    print(
        line_to_output.encode(file_encoding, errors="ignore").decode(),
        end=ending,
        file=file_descriptor,
    )


def cabrillo(self, file_encoding):
    """Generates Cabrillo file. Maybe."""
    # https://www.cqwpx.com/cabrillo.htm
    logger.debug("******Cabrillo*****")
    logger.debug("Station: %s", f"{self.station}")
    logger.debug("Contest: %s", f"{self.contest_settings}")
    now = datetime.datetime.now()
    date_time = now.strftime("%Y-%m-%d_%H-%M-%S")
    filename = (
        str(Path.home())
        + "/"
        + f"{self.station.get('Call', '').upper().replace('/','-')}_{cabrillo_name}_{date_time}.log"
    )
    logger.debug("%s", filename)
    log = self.database.fetch_all_contacts_asc()
    try:
        with open(filename, "w", encoding=file_encoding, newline="") as file_descriptor:
            output_cabrillo_line(
                "START-OF-LOG: 3.0",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CREATED-BY: Not1MM v{__version__}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CONTEST: {cabrillo_name}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.station.get("Club", ""):
                output_cabrillo_line(
                    f"CLUB: {self.station.get('Club', '').upper()}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"CALLSIGN: {self.station.get('Call','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"LOCATION: {self.station.get('ARRLSection', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-OPERATOR: {self.contest_settings.get('OperatorCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-ASSISTED: {self.contest_settings.get('AssistedCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-BAND: {self.contest_settings.get('BandCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            mode = self.contest_settings.get("ModeCategory", "")
            if mode in ["SSB+CW", "SSB+CW+DIGITAL"]:
                mode = "MIXED"
            output_cabrillo_line(
                f"CATEGORY-MODE: {mode}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-TRANSMITTER: {self.contest_settings.get('TransmitterCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            if self.contest_settings.get("OverlayCategory", "") != "N/A":
                output_cabrillo_line(
                    f"CATEGORY-OVERLAY: {self.contest_settings.get('OverlayCategory','')}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line(
                f"GRID-LOCATOR: {self.station.get('GridSquare','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"CATEGORY-POWER: {self.contest_settings.get('PowerCategory','')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )

            output_cabrillo_line(
                f"CLAIMED-SCORE: {calc_score(self)}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            ops = ""
            list_of_ops = self.database.get_ops()
            for op in list_of_ops:
                ops += f"{op.get('Operator', '')}, "
            if self.station.get("Call", "") not in ops:
                ops += f"@{self.station.get('Call','')}"
            else:
                ops = ops.rstrip(", ")
            output_cabrillo_line(
                f"OPERATORS: {ops}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"NAME: {self.station.get('Name', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS: {self.station.get('Street1', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-CITY: {self.station.get('City', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-STATE-PROVINCE: {self.station.get('State', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-POSTALCODE: {self.station.get('Zip', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"ADDRESS-COUNTRY: {self.station.get('Country', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            output_cabrillo_line(
                f"EMAIL: {self.station.get('Email', '')}",
                "\r\n",
                file_descriptor,
                file_encoding,
            )
            for contact in log:
                the_date_and_time = contact.get("TS", "")
                themode = contact.get("Mode", "")
                # if themode == "LSB" or themode == "USB":
                #    themode = "PH"

                match themode:
                    case "LSB" | "USB" | "SSB" | "FM" | "AM":
                        themode = "PH"
                    case "CW" | "CW-U" | "CW-L" | "CWR" | "CW-R":
                        themode = "CW"
                    # dont simplify digital mode names
                    # case "FT8"|"FT4"|"RTTY"|"PSK31"|"FSK441"|"MSK144"|"JT65"|"JT9"|"Q65"|"PKTUSB"|"PKTLSB":
                    #    themode = "DI"
                    # case _:
                    #    mode_test = "OTHER"
                # end match

                frequency = str(round(contact.get("Freq", "0"))).rjust(5)

                if contact.get("RoverLocation", ""):
                    location = f"{contact.get('RoverLocation').ljust(15).upper()}"
                elif self.contest_settings.get("SentExchange", ""):
                    location = f"{self.contest_settings.get('SentExchange', '').ljust(15).upper()}"
                else:
                    # user did not provide any location info,
                    # so insert a placeholder in cabrillo file
                    location = f"*              "

                loggeddate = the_date_and_time[:10]
                loggedtime = the_date_and_time[11:13] + the_date_and_time[14:16]
                output_cabrillo_line(
                    f"QSO: {frequency} {themode} {loggeddate} {loggedtime} "
                    f"{contact.get('StationPrefix', '').ljust(13)} "
                    f"{str(contact.get('SNT', '')).ljust(6)} "
                    f"{location} "
                    f"{contact.get('Call', '').ljust(13)} "
                    f"{str(contact.get('RCV', '')).ljust(6)} "
                    f"{str(contact.get('Exchange1', '')).ljust(15)}",
                    "\r\n",
                    file_descriptor,
                    file_encoding,
                )
            output_cabrillo_line("END-OF-LOG:", "\r\n", file_descriptor, file_encoding)
        self.show_message_box(f"Cabrillo saved to: {filename}")
    except IOError as exception:
        logger.critical("cabrillo: IO error: %s, writing to %s", exception, filename)
        self.show_message_box(f"Error saving Cabrillo: {exception} {filename}")
        return


def populate_history_info_line(self):
    result = self.database.fetch_call_history(self.callsign.text())
    if result:
        self.history_info.setText(
            f"{result.get('Call', '')}, {result.get('Exchange1', '')}"
        )
    else:
        self.history_info.setText("")


def check_call_history(self):
    """"""
    result = self.database.fetch_call_history(self.callsign.text())
    print(f"{result=}")
    if result:
        self.history_info.setText(f"{result.get('Exchange1','')}")


def process_esm(self, new_focused_widget=None, with_enter=False):
    """ESM State Machine"""

    # self.pref["run_state"]

    # -----===== Assigned F-Keys =====-----
    # self.esm_dict["CQ"]
    # self.esm_dict["EXCH"]
    # self.esm_dict["QRZ"]
    # self.esm_dict["AGN"]
    # self.esm_dict["HISCALL"]
    # self.esm_dict["MYCALL"]
    # self.esm_dict["QSOB4"]

    # ----==== text fields ====----
    # self.callsign
    # self.sent
    # self.receive
    # self.other_1
    # self.other_2

    if new_focused_widget is not None:
        self.current_widget = self.inputs_dict.get(new_focused_widget)

    # print(f"checking esm {self.current_widget=} {with_enter=} {self.pref.get("run_state")=}")

    for a_button in [
        self.esm_dict["CQ"],
        self.esm_dict["EXCH"],
        self.esm_dict["QRZ"],
        self.esm_dict["AGN"],
        self.esm_dict["HISCALL"],
        self.esm_dict["MYCALL"],
        self.esm_dict["QSOB4"],
    ]:
        if a_button is not None:
            self.restore_button_color(a_button)

    buttons_to_send = []

    if self.pref.get("run_state"):
        if self.current_widget == "callsign":
            if len(self.callsign.text()) < 3:
                self.make_button_green(self.esm_dict["CQ"])
                buttons_to_send.append(self.esm_dict["CQ"])
            elif len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["HISCALL"])
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["HISCALL"])
                buttons_to_send.append(self.esm_dict["EXCH"])

        elif self.current_widget == "other_1":
            if self.other_1.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["QRZ"])
                buttons_to_send.append(self.esm_dict["QRZ"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    self.process_function_key(button)
    else:
        if self.current_widget == "callsign":
            if len(self.callsign.text()) > 2:
                self.make_button_green(self.esm_dict["MYCALL"])
                buttons_to_send.append(self.esm_dict["MYCALL"])

        elif self.current_widget == "other_1":
            if self.other_1.text() == "":
                self.make_button_green(self.esm_dict["AGN"])
                buttons_to_send.append(self.esm_dict["AGN"])
            else:
                self.make_button_green(self.esm_dict["EXCH"])
                buttons_to_send.append(self.esm_dict["EXCH"])
                buttons_to_send.append("LOGIT")

        if with_enter is True and bool(len(buttons_to_send)):
            for button in buttons_to_send:
                if button:
                    if button == "LOGIT":
                        self.save_contact()
                        continue
                    self.process_function_key(button)


def get_mults(self):
    """Get mults for RTC XML"""


def just_points(self):
    """Get points for RTC XML"""
